from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from six.moves import range
from six.moves import zip
try:
    import numpy as np
    import reflex
    from pipeline_product import PipelineProduct
    import pipeline_display
    import os
    import re
    import matplotlib.gridspec as gridspec
    from matplotlib.figure import Figure
    from matplotlib.backends.backend_wxagg import FigureCanvasWxAgg
    import_success = True
    import wx
    import wx.grid
    from wx import xrc
    import wx.lib.mixins.listctrl as listmix
    from wx.lib.mixins.listctrl import CheckListCtrlMixin,ListCtrlAutoWidthMixin

except ImportError:
    import_success = False
    print("Error importing modules pyfits, wx, matplotlib, numpy")

#from copy import deepcopy as dc
from datetime import datetime
from matplotlib.ticker import MaxNLocator

class CheckListCtrl(wx.ListCtrl, CheckListCtrlMixin, ListCtrlAutoWidthMixin):
    def __init__(self, parent):
        wx.ListCtrl.__init__(self, parent, -1,
                             style=wx.LC_REPORT|wx.SUNKEN_BORDER)
        CheckListCtrlMixin.__init__(self)
        ListCtrlAutoWidthMixin.__init__(self)

class EditCheckListCtrl(wx.ListCtrl, CheckListCtrlMixin, ListCtrlAutoWidthMixin,
                        listmix.TextEditMixin):
    def __init__(self, parent):
        wx.ListCtrl.__init__(self, parent, -1,
                             style=wx.LC_REPORT|wx.SUNKEN_BORDER)
        listmix.TextEditMixin.__init__(self)
        CheckListCtrlMixin.__init__(self)
        ListCtrlAutoWidthMixin.__init__(self)
        self.Bind(wx.EVT_LIST_BEGIN_LABEL_EDIT, self.onBeginLabelEdit)

    def onBeginLabelEdit(self, event):

        # To make the first 2 columns non-editable
        if event.m_col < 2:
            event.Veto()
        else:
            event.Skip()

class DataPlotterManager:

    recipe_name = "espda_fit_line"

    """ General methods """

    def addSubplots(self, figure):
        self.gs = gridspec.GridSpec(42, 6)
        self.figure = figure
        self.p = Plot()
        self.p.sub_b = self.p.sub(self.figure, self.gs[37:,0:1], xlabel=None,
                                  ylabel=None)


    def plotProductsGraphics(self):
        self.error_s = True
        self.fit_s = True
        #self.edit_string = ""
        #self.edit_name = "dummy.txt"
        self.vals = []
        self.plotSpec()


    def plotSpec(self):
        self.plotSpecWhole()
        self.plotSpecSyst()


    def readFitsData(self, fitsFiles):
        frames = dict()
        for frame in fitsFiles:
            if frame == '':
                continue
            frames[frame.category] = PipelineProduct(frame)

        try:
            self.rspec_voigt = Data(frames["RSPEC_VOIGT"])
            self.sline_iden = Data(frames["SLINE_VOIGT"], y="PEAK")
            self.fline_sel = Data(frames["SLINE_VOIGT"], y="PEAK")
        except:
            self.rspec_voigt = Data(frames["RSPEC_GUESS"])
            self.sline_iden = Data(frames["SLINE_GUESS"], y="PEAK")
            self.fline_sel = Data(frames["SLINE_GUESS"], y="PEAK")
            
        
        self.qso_ion = Data(frames["QSO_ION"], y="OSCS")
        self.ion_wavel = {i.strip(): w \
                          for (i, w) in zip(self.qso_ion.l, self.qso_ion.x)}
        #self.fline_sel = dc(self.sline_iden)
        self.syst_num = len(set(self.sline_iden.s))
        self.systid = sorted(list(set(self.sline_iden.s)))
        self.systid_sel = str(self.systid[0])

        sub = re.sub("FLINE_IDEN.fits", "", frames["FLINE_IDEN"].fits_file.name)
        path = "/".join(sub.split("/")[:-2])+"/"
        time = datetime.now().time().isoformat()#sub.split("/")[-2]
        prev = [e[29:-4] for e in os.listdir(path) if e[:28]=='edit_'+time]
        try:
            self.edit_name = path+"edit_"+time+"_"+str(int(prev[-1])+1)+".txt"
        except:
            self.edit_name = path+"edit_"+time+"_1.txt"


    def setInteractiveParameters(self):
        paramList      = list()
        paramList.append(reflex.RecipeParameter(recipe = self.recipe_name,
            displayName = "hwidth",    group = "Line parameters", description =
            "Halfwidth for line grouping [nm]: [2.5e-3, 10.0]"))
        paramList.append(reflex.RecipeParameter(recipe = self.recipe_name,
            displayName = "par-range", group = "Line parameters", description =
            "Rules to set the range of line parameters\n" \
            "Sample syntax for par-range: '1e-4,10.0,17.0,2.0,50.0,0.0,0.0', " \
            "meaning:\n" \
            "- redshift half-range: 1.0e-4\n" \
            "- minimum/maximum logarithimic column density: 10.0/17.0\n" \
            "- minimum/maximum Doppler broadening: 2.0/50.0\n" \
            "- minimum/maximum turbulence broadening: 0.0/0.0"))
        paramList.append(reflex.RecipeParameter(recipe = self.recipe_name,
            displayName = "edit-file", group = "Line parameters", description =
            "File with instructions to edit line table"))
        paramList.append(reflex.RecipeParameter(recipe = self.recipe_name,
            displayName = "fit-flag",  group = "Fitting",         description =
            "Flag to fit the lines"))
        paramList.append(reflex.RecipeParameter(recipe = self.recipe_name,
            displayName = "iter-max",  group = "Fitting",         description =
            "Maximum number of iterations: [1, 1e3]"))
        paramList.append(reflex.RecipeParameter(recipe = self.recipe_name,
            displayName = "grp-sel",   group = "Fitting",         description =
            "Group of lines selected for fitting: [0, number of groups]"))
        paramList.append(reflex.RecipeParameter(recipe = self.recipe_name,
            displayName="model-resol", group = "Fitting",         description =
            "Resolution of the instrument model [1.0e3, 1.0e6]"))
        return paramList


    """ Specific methods """

    def addLine(self, point):
        self.ltm = lineTabManager
        self.ipm = ionPanelManager

        cond = np.where([point.inaxes.get_yaxis()==s.axes.get_yaxis() \
                         for s in self.p.sub_s])[0][0]
        val = self.rspec_voigt.get_xy(point.xdata, point.ydata, self.v_s[cond])
        wavel = val[0]
        lineid = self.lineid_sel[cond]
        ion_wavel = self.ion_wavel[lineid]
        if val in self.vals:
            self.vals.remove(val)
        else:
            self.vals.append(val)
            self.ltm.onAddLine(wavel, lineid, ion_wavel)

        """
        for w in wxApp.shownParamWidgets:
            if w[1].parameter.displayName == 'add' and self.first_call:
                self.add_old = w[1].paramChangeCtrl.GetValue()
                self.first_call = False
        """
        self.plotSpecAdd()
        """
        val_str = ""
        for v in self.vals:
            val_str += str(v[0])+"-"+str(v[1])+","
        """
        new_params = list()
        """
        new_params.append(reflex.RecipeParameter(
            recipe=self.recipe_name, displayName="add",
            value=self.add_old+val_str))
        """
        return new_params

    def editString(self, edit):
        """
        if col == 2:
            col_mod = 0
        elif col == 3:
            col_mod = 2
        elif col == 4:
            col_mod = 1
        else:
            col_mod = col
        self.edit_string = self.edit_string+str(order)+":"+str(col_mod)+":"\
                           +edit+","

        edit_old = str()
        """
        for widget in wxApp.shownParamWidgets:
            if widget[1].parameter.displayName == "edit-file":
                widget[1].paramChangeCtrl.SetValue(self.edit_name)
        edit_file = open(self.edit_name, "w")
        edit_file.write(edit)
        edit_file.close()


    def plotSpecAdd(self):
        x = [v[0] for v in self.vals]
        y = [v[1] for v in self.vals]
        self.p.sub_w.scatter(x, y, marker='+', s=300, linewidth=1.5, color='r')
        for s, c in zip(self.p.sub_s, self.cen_s):
            s.scatter((1-c/x)*299792.458, y, marker='+', s=300,
                      linewidth=1.5, color='r')
        try:
            wxApp.figureCanvas.draw()
        except:
            pass


    def plotSpecSyst(self, refresh_lineid=True, xlim=(-200,200),
                     ylim=(-0.5,1.5)):
        try:
            self.fline_sel_a.remove()
        except:
            pass
        self.fline_sel.select("SYSTID", self.systid_sel)
        self.fline_sel_a = self.fline_sel.scatter(
            self.p.sub_w, norm=True, marker='+', s=40, linewidth=1.5, color='g',
            zorder=10, label="selected")

        self.p.sub_w.legend(prop={'size': 10}, loc=3)

        if refresh_lineid:
            self.lineid_sel = list(np.unique(
                self.fline_sel.l[self.fline_sel.sel]))
        self.systz_sel = np.mean(self.fline_sel.z[self.fline_sel.sel])
        self.wavel_sel = np.array([self.ion_wavel[i.strip()] \
                                   for i in self.lineid_sel])
        ns = len(self.wavel_sel)
        nr = min(ns, 4)
        nc = min((ns-1)//4+1, 2)

        if ns > 8:
            print("Plotting only 8 panels")

        tooltip_s = ("Detail of the selected absorption system, split into "
                     "subplots of identified transitions. Check `error` to "
                     "hide overplot the error spectrum. Check `fit/resid.` to "
                     "hide overplot the spectrum of fitted lines and fitting "
                     "residuals. Check entries in the `Transitions` window to "
                     "display other transitions and look for possible lines to "
                     "add to the sistems. Middle-click on subplots to add "
                     "lines. Added lines will inherit the transition from the "
                     "subplot (as denoted by its label). Adjust the parameters "
                     "of existing and added lines in the `Lines` table, "
                     "following instructions in the tutorial.")
        self.p.sub_s = []
        self.cen_s = []
        self.rspec_voigt_a = []
        self.v_s = []
        for i, (l, w) \
            in enumerate(zip(self.lineid_sel, self.wavel_sel)):
            te = 12+(i%nr)*24//nr
            be = 12+(i%nr+1)*24//nr
            le = i//nr*6//nc
            re = (i//nr+1)*6//nc
            s = self.p.sub(self.figure, self.gs[te:be,le:re],
                           xlabel=r"Velocity difference [km/s]",
                           ylabel=None, tooltip=tooltip_s)

            if i != 0:
                s.get_shared_x_axes().join(s, self.p.sub_s[0])
                s.get_shared_y_axes().join(s, self.p.sub_s[0])
            if i != nr-1 and i != ns-1:
                s.set_xticklabels([])
                s.set_xlabel("")
            if i >= nr:
                s.set_yticklabels([])
                s.set_ylabel("")
            if xlim != None:
                s.set_xlim(xlim)
            if ylim != None:
                s.set_ylim(ylim)

            s.yaxis.set_major_locator(MaxNLocator(prune='both'))
            c = (1+self.systz_sel)*w
            self.cen_s.append(c)
            self.p.sub_s.append(s)
            a, = self.rspec_voigt.step(
                s, xlim=xlim, error=self.error_s, fit=self.fit_s,
                resid=self.fit_s, norm=True, cen=c, label=l)
            self.rspec_voigt_a.append(a)
            self.v_s.append(self.rspec_voigt.v)
            self.sline_iden.scatter(s, norm=True, cen=c, marker='+', s=40,
                                    linewidth=1.5, color='lightgray', zorder=9)
            self.fline_sel.scatter(s, norm=True, cen=c, marker='+', s=40,
                                   linewidth=1.5, color='g', zorder=10)
            s.text(0.05, 0.5, l, ha='left', va='center',
                   transform=s.transAxes)

        try:
            self.plotWidgetClick()
        except:
            pass
        self.plotSpecAdd()
        #wxApp.figureCanvas.draw()

        for s in self.p.sub_s:
            s.callbacks.connect(
                'xlim_changed',
                lambda e: self.p.rectangles(e, self.p.sub_s[0].get_ylim(),
                                            self.cen_s))
        self.p.rectangles(ylim=self.p.sub_s[0].get_ylim(), cen=self.cen_s)


    def plotSpecWhole(self):
        tooltip_w = ("Continuum-fitted spectrum (frame category: RSPEC_CONT) "
                     "with identified (frame category: FLINE_IDEN) and "
                     "non-identified lines (frame category: FLINE_REJ).")
        self.p.sub_w = self.p.sub(self.figure, self.gs[:9,:],
                                  title="Identified systems (%i)" \
                                  % self.syst_num, ylabel="Normalized flux",
                                  tooltip=tooltip_w)
        self.rspec_voigt.step(self.p.sub_w, norm=True)
        self.sline_iden.scatter(self.p.sub_w, norm=True, marker='+', s=40,
                                linewidth=1.5, color='lightgray', zorder=9)

    def plotWidgets(self):
        widgets = list()

        # Clickable plot
        self.first_call = True
        self.click = []
        for s in self.p.sub_s:
            click = InteractiveClickableSubplot(s, self.addLine)
            self.click.append(click)
            widgets.append(click)

        # Check buttons
        self.toggle = InteractiveCheckButtons(
            self.p.sub_b, self.togglePlotElem, ('error', 'fit/resid.'), (1, 1))
        widgets.append(self.toggle)

        return widgets

    def togglePlotElem(self, sel):
        if sel == 'error':
            self.error_s = not(self.error_s)
        elif sel == 'fit/resid.':
            self.fit_s = not(self.fit_s)
        xlim = self.p.sub_s[0].get_xlim()
        ylim = self.p.sub_s[0].get_ylim()
        """
        for s in self.p.sub_s:
            d = self.rspec_voigt
            if self.error_s:
                self.error_p = s.plot(d.x[d.sel]*d.v[d.sel],
                                      d.e[d.sel]*d.n[d.sel],
                                      label=d.label+" "+"error")
            else:
                try:
                    del self.error_p
                except:
                    pass
            if self.fit_s:
                self.fit_p = s.plot(d.x[d.sel]*d.v[d.sel],
                                    d.f[d.sel]*d.n[d.sel],
                                    label=d.label+" "+"fit")
                self.resid_p = s.plot(d.x[d.sel]*d.v[d.sel],
                                      d.r[d.sel]*d.n[d.sel],
                                      label=d.label+" "+"resid.")
            else:
                try:
                    del self.fit_p
                    del self.resid_p
                except:
                    pass
        """
        for s in self.p.sub_s:
            s.cla()
        self.plotSpecSyst(xlim=xlim, ylim=ylim, refresh_lineid=False)


class IonPanelManager(wx.Frame):
    def __init__(self, parent, id, title):
        wx.Frame.__init__(self, parent, id, title, size = (454, 300),
            pos = (0, 0))

        self.dpm = dataPlotManager
        self.selected = []

        panel = wx.Panel(self, -1)
        vbox = wx.BoxSizer(wx.VERTICAL)

        hbox1 = wx.BoxSizer(wx.HORIZONTAL)
        top_panel = wx.Panel(panel, -1)
        self.cl = CheckListCtrl(top_panel)
        self.cl.InsertColumn(0, "", width = 23)
        self.cl.InsertColumn(1, "Line ID", width = 100)
        self.cl.InsertColumn(2, "Rest-frame", width = 100)
        self.cl.InsertColumn(3, "Redshifted", width = 100)
        self.cl.InsertColumn(4, "Osc. strength", width = 100)
        hbox1.Add(self.cl, 1, wx.EXPAND)
        top_panel.SetSizer(hbox1)
        vbox.Add(top_panel, 1, wx.EXPAND | wx.ALL, 5)
        panel.SetSizer(vbox)

        wavel_min = np.min(self.dpm.rspec_voigt.x)
        wavel_max = np.max(self.dpm.rspec_voigt.x)
        ion = self.dpm.qso_ion
        sel = np.where(np.logical_and(ion.x*(1+self.dpm.systz_sel)>wavel_min,
                                      ion.x*(1+self.dpm.systz_sel)<wavel_max))
        for l, x, y \
            in zip(ion.l[sel], ion.x[sel], ion.y[sel]):
            try:
                index = self.cl.InsertItem(sys.maxsize, "")
                self.cl.SetItem(index, 1, str(l))
                self.cl.SetItem(index, 2, str(x))
                self.cl.SetItem(index, 3, str(x*(1+self.dpm.systz_sel)))
                self.cl.SetItem(index, 4, str(y))
            except:  # Python 2
                index = self.cl.InsertStringItem(sys.maxsize, "")
                self.cl.SetStringItem(index, 1, str(l))
                self.cl.SetStringItem(index, 2, str(x))
                self.cl.SetStringItem(index, 3, str(x*(1+self.dpm.systz_sel)))
                self.cl.SetStringItem(index, 4, str(y))

            if l.strip() in self.dpm.lineid_sel:
                self.cl.CheckItem(index)
                self.selected.append(index)
        self.Show()
        self.cl.OnCheckItem = self.onCheckItem


    def onCheckItem(self, index, flag):
        i = str(self.cl.GetItem(index, 1).GetText()).strip()
        if i in self.dpm.lineid_sel:
            self.dpm.lineid_sel.remove(i)
            self.selected.remove(index)
        else:
            self.dpm.lineid_sel.append(i)
            self.selected.append(index)
        if len(self.dpm.lineid_sel) > 8:
            self.cl.CheckItem(self.selected[0], check=False)
            self.selected = self.selected[-8:]
            self.dpm.lineid_sel = self.dpm.lineid_sel[-8:]
        for s in self.dpm.p.sub_s:
            wxApp.figure.delaxes(s)
        self.dpm.plotSpecSyst(refresh_lineid=False)
        self.dpm.plotWidgets()
        wxApp.figureCanvas.draw()

    """
    def updateLineList(self):
        self.lineid_list = list()
        self.ion_wavel   = list()
        self.sub_cen     = list()
        last = -1
        while True:
            index = self.list.GetNextItem(last)
            if index == -1:
                break
            else:
                last = index
                lineid = str(self.list.GetItem(index, 1).GetText()).strip()
                if self.list.IsChecked(index):
                    self.lineid_list.append("{:9s}".format(lineid))
                    self.ion_wavel.append(float(self.list.GetItem(index, 2).\
                                                GetText()))
                    self.sub_cen.append(float(self.list.GetItem(index, 3).\
                                              GetText()))
                    mode = True
                    color = "black"
                    if self.index == index:
                        self.check_line = float(self.list.GetItem(index, 3).\
                                                  GetText())
                else:
                    mode = True
                    color = "gray"
            lineTabManager.refresh(lineid, mode, color)
        self.sub_num = len(self.sub_cen)

    def updateLinePlot(self):
        if self.check_line not in self.show_line:
            self.show_line = np.array(self.show_line)
            self.show_line = np.append(self.show_line, self.check_line)
        else:
            w = np.where(
                self.show_line != self.check_line)[0]
            self.show_line = np.array(self.show_line[w])
        dataPlotManager.show_line = self.show_line
    """

class LineTabManager(wx.Frame):
    cols = {'all':[6,9,12], 'linez':[6], 'colden':[9], 'therb':[12],
            'turbb':[15]}
    tags = {'all':['z','N','b'], 'linez':['z'], 'colden':['N'], 'therb':['b'],
            'turbb':['btur']}
    vals = {'all': ['linez_v', 'colden_v', 'therb_v'], 'linez':['linez_v'],
            'colden':['colden_v'], 'therb':['therb_v'], 'turbb':['turbb_v']}

    def __init__(self, parent, id, title):
        wx.Frame.__init__(self, parent, id, title, size = (1216, 300),
            pos = (0, 600))
        self.panel = wx.Panel(self)

        self.dpm = dataPlotManager
        self.ipm = ionPanelManager
        self.edit = ""

        #self.replot = True

        self.cl = EditCheckListCtrl(self.panel)
        self.cl.InsertColumn(0,  "", width = 23)
        self.cl.InsertColumn(1,  "#", width = 52)
        self.cl.InsertColumn(2,  "Line ID", width = 80)
        self.cl.InsertColumn(3,  "Wavel.", width = 80)
        self.cl.InsertColumn(4,  "Redshift", width = 80)
        self.cl.InsertColumn(5,  "error", width = 80)
        self.cl.InsertColumn(6,  "constr.", width = 80)
        self.cl.InsertColumn(7,  "Col. d.", width = 80)
        self.cl.InsertColumn(8,  "error", width = 80)
        self.cl.InsertColumn(9,  "constr.", width = 80)
        self.cl.InsertColumn(10, "Therm. br.", width = 80)
        self.cl.InsertColumn(11, "error", width = 80)
        self.cl.InsertColumn(12, "constr.", width = 80)
        self.cl.InsertColumn(13, "Turb. br.", width = 80)
        self.cl.InsertColumn(14, "error", width = 80)
        self.cl.InsertColumn(15, "constr.", width = 80)
        self.linenum = 0
        frame = self.dpm.sline_iden.frame
        hdulist   = frame.hdulist()
        order     = frame.readTableColumn(1, "ORDER")
        lineid    = frame.readTableColumn(1, "LINEID")
        wavel     = frame.readTableColumn(1, "WAVEL")
        linez     = frame.readTableColumn(1, "LINEZ")
        linezerr  = frame.readTableColumn(1, "LINEZERR")
        linezcon  = frame.readTableColumn(1, "LINEZCON")
        colden    = frame.readTableColumn(1, "COLDEN")
        coldenerr = frame.readTableColumn(1, "COLDENERR")
        coldencon = frame.readTableColumn(1, "COLDENCON")
        therb     = frame.readTableColumn(1, "THERB")
        therberr  = frame.readTableColumn(1, "THERBERR")
        therbcon  = frame.readTableColumn(1, "THERBCON")
        turbb     = frame.readTableColumn(1, "TURBB")
        turbberr  = frame.readTableColumn(1, "TURBBERR")
        turbbcon  = frame.readTableColumn(1, "TURBBCON")
        self.linenum = len(lineid)
        self.linemax = max(order)
        for i in range(0, self.linenum):
            try:
                index = self.cl.InsertItem(sys.maxsize, "")
                self.cl.SetItem(index, 1,  str(order[i]))
                self.cl.SetItem(index, 2,  lineid[i])
                self.cl.SetItem(index, 3,  "{:4.4f}".format(wavel[i]))
                self.cl.SetItem(index, 4,  "{:4.6f}".format(linez[i]))
                self.cl.SetItem(index, 5,
                                        "{:4.6f}".format(linezerr[i]))
                self.cl.SetItem(index, 6,  linezcon[i])
                self.cl.SetItem(index, 7,  "{:4.4f}".format(colden[i]))
                self.cl.SetItem(index, 8,
                                        "{:4.4f}".format(coldenerr[i]))
                self.cl.SetItem(index, 9,  coldencon[i])
                self.cl.SetItem(index, 10, "{:4.4f}".format(therb[i]))
                self.cl.SetItem(index, 11,
                                        "{:4.4f}".format(therberr[i]))
                self.cl.SetItem(index, 12, therbcon[i])
                self.cl.SetItem(index, 13, "{:4.4f}".format(turbb[i]))
                self.cl.SetItem(index, 14,
                                       "{:4.4f}".format(turbberr[i]))
                self.cl.SetItem(index, 15, turbbcon[i])
            except:  # Python 2
                index = self.cl.InsertStringItem(sys.maxsize, "")
                self.cl.SetStringItem(index, 1,  str(order[i]))
                self.cl.SetStringItem(index, 2,  lineid[i])
                self.cl.SetStringItem(index, 3,  "{:4.4f}".format(wavel[i]))
                self.cl.SetStringItem(index, 4,  "{:4.6f}".format(linez[i]))
                self.cl.SetStringItem(index, 5,
                                        "{:4.6f}".format(linezerr[i]))
                self.cl.SetStringItem(index, 6,  linezcon[i])
                self.cl.SetStringItem(index, 7,  "{:4.4f}".format(colden[i]))
                self.cl.SetStringItem(index, 8,
                                        "{:4.4f}".format(coldenerr[i]))
                self.cl.SetStringItem(index, 9,  coldencon[i])
                self.cl.SetStringItem(index, 10, "{:4.4f}".format(therb[i]))
                self.cl.SetStringItem(index, 11,
                                        "{:4.4f}".format(therberr[i]))
                self.cl.SetStringItem(index, 12, therbcon[i])
                self.cl.SetStringItem(index, 13, "{:4.4f}".format(turbb[i]))
                self.cl.SetStringItem(index, 14,
                                       "{:4.4f}".format(turbberr[i]))
                self.cl.SetStringItem(index, 15, turbbcon[i])
            self.cl.CheckItem(index)
        self.count = np.max([int(c[-3:].strip() or 0) for c in linezcon if '-' in c])

        """
        if len(dataPlotManager.add_wavel) > 0:
            self.onAddLine()

        self.disc = [1] * self.linenum
        #self.disc_add = list()
        self.cl.Bind(wx.EVT_LIST_END_LABEL_EDIT, self.onEdit)
        self.Bind(wx.EVT_LIST_ITEM_SELECTED, self.onSelItem)
        self.cl.OnCheckItem = self.onCheckItem
        """
        self.sel = []
        self.cl.Bind(wx.EVT_LIST_END_LABEL_EDIT, self.onEdit)
        #self.cl.Bind(wx.EVT_LIST_INSERT_ITEM, self.onEdit)
        self.Bind(wx.EVT_LIST_ITEM_SELECTED, self.onSelItem)

        vsizer = wx.BoxSizer(wx.VERTICAL)
        vsizer.Add(self.cl, 1, wx.EXPAND | wx.ALL, 5)
        self.value_list = []
        hsizer = wx.BoxSizer(wx.HORIZONTAL)
        undo = wx.Button(self.panel, label = "Undo", size = (50, 25))
        hsizer.Add(undo, 0, wx.CENTER | wx.LEFT,    5)
        undo.Bind(wx.EVT_BUTTON, self.onUndo)
        hsizer.Add(self.threeButtons(25, "all"),    0)
        hsizer.Add(self.threeButtons(5, "linez"),  0)
        hsizer.Add(self.threeButtons(85, "colden"), 0)
        hsizer.Add(self.threeButtons(85, "therb"),  0)
        hsizer.Add(self.threeButtons(85, "turbb"),  0)
        vsizer.Add(hsizer,  0, wx.BOTTOM, 5)
        self.panel.SetSizer(vsizer)
        self.Show()


    def onAddLine(self, wavel, lineid, ion_wavel):
        linez = wavel / ion_wavel - 1
        try:
            index = self.cl.InsertItem(sys.maxsize, "")
            self.cl.SetItem(index, 1, str(self.cl.GetItemCount()))
            self.cl.SetItem(index, 2, lineid)
            self.cl.SetItem(index, 3, "{:4.4f}".format(wavel))
            self.cl.SetItem(index, 4, "{:4.6f}".format(linez))
            self.cl.SetItem(index, 6, "V")
            self.cl.SetItem(index, 7, str(13.0))
            self.cl.SetItem(index, 9, "V")
            self.cl.SetItem(index, 10, str(10.0))
            self.cl.SetItem(index, 12, "V")
            self.cl.SetItem(index, 13, str(0.0))
            self.cl.SetItem(index, 15, "F")
        except:  # Python 2
            index = self.cl.InsertStringItem(sys.maxsize, "")
            self.cl.SetStringItem(index, 1, str(self.cl.GetItemCount()))
            self.cl.SetStringItem(index, 2, lineid)
            self.cl.SetStringItem(index, 3, "{:4.4f}".format(wavel))
            self.cl.SetStringItem(index, 4, "{:4.6f}".format(linez))
            self.cl.SetStringItem(index, 6, "V")
            self.cl.SetStringItem(index, 7, str(13.0))
            self.cl.SetStringItem(index, 9, "V")
            self.cl.SetStringItem(index, 10, str(10.0))
            self.cl.SetStringItem(index, 12, "V")
            self.cl.SetStringItem(index, 13, str(0.0))
            self.cl.SetStringItem(index, 15, "F")

        self.cl.CheckItem(index)
        """
        order = int(self.cl.GetItem(index, 1).GetText())
        #dataPlotManager.editString(order, 1, str(self.linemax + len(add_wavel)))
        dataPlotManager.editString(order, 2, lineid.strip())
        dataPlotManager.editString(order, 3, "{:4.4f}".format(wavel))
        dataPlotManager.editString(order, 4, "{:4.6f}".format(linez))
        dataPlotManager.editString(order, 6, "V")
        dataPlotManager.editString(order, 7, str(13.0))
        dataPlotManager.editString(order, 9, "V")
        dataPlotManager.editString(order, 10, str(10.0))
        dataPlotManager.editString(order, 12, "V")
        dataPlotManager.editString(order, 13, str(0.0))
        dataPlotManager.editString(order, 15, "F")
        """
        self.onEdit(None)

    def onUndo(self, event):
        for s,u in zip(self.sel, self.undo):
            for c, v in enumerate(u):
                try:
                    self.cl.SetItem(s, c, v)
                except:  # Python 2
                    self.cl.SetStringItem(s, c, v)
        self.last_edit = ""

    """
        hdulist   = self.sline.hdulist()
        order     = self.sline.readTableColumn(1, "ORDER")
        lineid    = self.sline.readTableColumn(1, "LINEID")
        wavel     = self.sline.readTableColumn(1, "WAVEL")
        linez     = self.sline.readTableColumn(1, "LINEZ")
        linezerr  = self.sline.readTableColumn(1, "LINEZERR")
        linezcon  = self.sline.readTableColumn(1, "LINEZCON")
        colden    = self.sline.readTableColumn(1, "COLDEN")
        coldenerr = self.sline.readTableColumn(1, "COLDENERR")
        coldencon = self.sline.readTableColumn(1, "COLDENCON")
        therb     = self.sline.readTableColumn(1, "THERB")
        therberr  = self.sline.readTableColumn(1, "THERBERR")
        therbcon  = self.sline.readTableColumn(1, "THERBCON")
        turbb     = self.sline.readTableColumn(1, "TURBB")
        turbberr  = self.sline.readTableColumn(1, "TURBBERR")
        turbbcon  = self.sline.readTableColumn(1, "TURBBCON")
        self.linenum = len(lineid)
        i = 0
        last = -1
        while True:
            index = self.cl.GetNextItem(last)
            if index == -1:
                break
            else:
                last = index
                try:
                    self.cl.SetItem(index, 1, str(order[i]))
                    self.cl.SetItem(index, 2, lineid[i])
                    self.cl.SetItem(index, 3,
                                            "{:4.4f}".format(wavel[i]))
                    self.cl.SetItem(index, 4,
                                            "{:4.6f}".format(linez[i]))
                    self.cl.SetItem(index, 5,
                                            "{:4.6f}".ormat(linezerr[i]))
                    self.cl.SetItem(index, 6, linezcon[i])
                    self.cl.SetItem(index, 7,
                                            "{:4.4f}".format(colden[i]))
                    self.cl.SetItem(index, 8,
                                            "{:4.4f}".ormat(coldenerr[i]))
                    self.cl.SetItem(index, 9, coldencon[i])
                    self.cl.SetItem(index, 10,
                                            "{:4.4f}".format(therb[i]))
                    self.cl.SetItem(index, 11,
                                            "{:4.4f}".format(therberr[i]))
                    self.cl.SetItem(index, 12, therbcon[i])
                    self.cl.SetItem(index, 13,
                                            "{:4.4f}".format(turbb[i]))
                    self.cl.SetItem(index, 14,
                                            "{:4.4f}".format(turbberr[i]))
                    self.cl.SetItem(index, 15, turbbcon[i])
                    i = i + 1
                except:
                    self.cl.DeleteItem(index);
        self.onSelLine()
        self.add_line = list()
        self.add_wavel = list()
        self.add_norm = list()
        self.show_line = list()

        #dataPlotManager.readFitsData(wxApp.all_fitsFiles)
        dataPlotManager.setPlotablePar()
        dataPlotManager.setPlotableData()
        dataPlotManager.plotProductsGraphics()
        wxApp.figureCanvas.draw()
    """

    def threeButtons(self, offset, param):
        bsizer = wx.BoxSizer(wx.HORIZONTAL)
        setattr(self, param+'_v', wx.TextCtrl(self.panel, size=(80, 25)))
        bsizer.Add(getattr(self, param+'_v'), 0, wx.CENTER|wx.LEFT, offset)
        bsizer.Add(self.oneButton("V", param), 0, wx.CENTER|wx.LEFT, 0)
        bsizer.Add(self.oneButton("F", param), 0, wx.CENTER|wx.LEFT, 0)
        bsizer.Add(self.oneButton("C", param), 0, wx.CENTER|wx.LEFT, 0)
        return bsizer

    def oneButton(self, label, param):
        button = wx.Button(self.panel, label=label, size=(25, 25))
        button.Bind(wx.EVT_BUTTON, lambda e: self.onButton(e, label, param))
        return button

    def onButton(self, event, label, param):
        """
        try:
            self.edit += self.last_edit
        except:
            pass
        self.last_edit = ""
        """
        self.undo = []
        if label == 'C':
            self.count += 1
        for s in self.sel:
            self.undo.append([self.cl.GetItem(s, c).GetText() \
                              for c in range(15)])
            order = self.cl.GetItem(s, 1).GetText()
            for c,t,v in zip(self.cols[param], self.tags[param],
                             self.vals[param]):
                value = getattr(self, v).GetValue()
                if label == 'C':
                    if c == 6:
                        wavel = (1+float(value))\
                                *float(self.cl.GetItem(s, 3).GetText())/\
                                (1+float(self.cl.GetItem(s, 4).GetText()))
                    constr = self.all_v.GetValue()+'-'+t\
                             +'-{:03d}'.format(self.count)
                    try:
                        self.cl.SetItem(s, c-2, value)
                        self.cl.SetItem(s, c, constr)
                    except:  # Python 2
                        self.cl.SetStringItem(s, c-2, value)
                        self.cl.SetStringItem(s, c, constr)
                    if c == 6:
                        try:
                            self.cl.SetItem(s, 3, '%4.4f' % wavel)
                        except:  # Python 2
                            self.cl.SetStringItem(s, 3, '%4.4f' % wavel)

                else:
                    try:
                        self.cl.SetItem(s, c, label)
                    except:  # Python 2
                        self.cl.SetStringItem(s, c, label)

                    #self.last_edit += order+":%i:%s," % (c, label)
        self.onEdit(event)


    def onEdit(self, event, iter=True):
        count = self.cl.GetItemCount()
        cols = self.cl.GetColumnCount()
        try:
            row_e = event.m_itemIndex
            col_e = event.GetColumn()
        except:
            row_e = -1
            col_e = -1
        self.edit = ""
        for row in range(count):
            for col in [-2, 0, 1, 2, 3, 4, 6, 7, 9, 10, 12, 13, 15]:
                if row == row_e and col == col_e:
                    label = event.GetLabel()
                elif col == -2:
                    label = self.dpm.systid_sel
                elif col == -1:
                    label = ""
                else:
                    item = self.cl.GetItem(itemId=row, col=col)
                    label = item.GetText()
                if col == 2:
                    col_mod = 0
                elif col == 3:
                    col_mod = 2
                elif col == 4:
                    col_mod = 1
                elif col == 0:
                    col_mod = 16
                else:
                    col_mod = col
                if label != "":
                    self.edit += self.cl.GetItem(row, 1).GetText()\
                                 +":"+str(col_mod)+":"+label+","

        #if iter:
        #    self.onEdit(event, iter=False)
        self.dpm.editString(self.edit)
        """
        try:
            self.edit += self.last_edit
        except:
            pass
        order = int(self.cl.GetItem(event.m_itemIndex, 1).GetText())
        col = event.GetColumn()
        label = event.GetLabel()
        print(col, label)
        self.last_edit = order+":%i:%s" % (col, label)
        print(self.last_edit)
        """


    def onSelItem(self, event):
        self.sel = []
        self.sel.append(self.cl.GetFirstSelected())
        while self.sel[-1] != -1:
            self.sel.append(self.cl.GetNextSelected(self.sel[-1]))
        self.sel = self.sel[:-1]
        self.all_v.SetValue(
            self.cl.GetItem(self.sel[-1], 2).GetText().strip()[:-5])
        self.linez_v.SetValue(self.cl.GetItem(self.sel[-1], 4).GetText())
        self.colden_v.SetValue(self.cl.GetItem(self.sel[-1], 7).GetText())
        self.therb_v.SetValue(self.cl.GetItem(self.sel[-1], 10).GetText())
        self.turbb_v.SetValue(self.cl.GetItem(self.sel[-1], 13).GetText())

    """
    def refresh(self, lineid, mode=True, color="black"):
        last = -1
        self.replot = False
        while True:
            index = self.cl.GetNextItem(last)
            if index == -1:
                break
            else:
                last = index
                self.lineid = str(self.cl.GetItem(index, 2).GetText()).strip()
                if lineid == self.lineid:
                    self.cl.CheckItem(index, mode)  # Very slow!
                    if mode is True:
                        self.cl.SetItemTextColour(index, color)
                    else:
                        self.cl.SetItemTextColour(index, color)
        self.replot = True


    def updateEditString(self):
        self.lineTab = ""
        last = -1
        while True:
            row = self.cl.GetNextItem(last)
            if row == -1:
                break
            else:
                last = row
                for col in range(0, 15):
                    #row = event.m_itemIndex
                    #edit = event.GetText()
                    data = str(self.cl.GetItem(row, col).GetText()).strip()
                    self.lineTab = self.lineTab+str(row)+":"+str(col)+":"+data\
                                   +","
    def onVall(self, event):
        self.onButton(2, self.value_list[0].GetValue())
        self.onButton(6, "V")
        self.onButton(9, "V")
        self.onButton(12, "V")
        #self.onButton(15, "V")

    def onVlinez(self, event):
        self.onButton(6, "V")

    def onVcolden(self, event):
        self.onButton(9, "V")

    def onVtherb(self, event):
        self.onButton(12, "V")

    def onVturbb(self, event):
        self.onButton(15, "V")

    def onFall(self, event):
        self.onButton(2, self.value_list[0].GetValue())
        self.onButton(6, "F")
        self.onButton(9, "F")
        self.onButton(12, "F")
        #self.onButton(15, "F")

    def onFlinez(self, event):
        self.onButton(6, "F")

    def onFcolden(self, event):
        self.onButton(9, "F")

    def onFtherb(self, event):
        self.onButton(12, "F")

    def onFturbb(self, event):
        self.onButton(15, "F")

    def onCall(self, event):
        lineid = self.value_list[0].GetValue()
        con = self.con_list[0]
        value = list()
        if con == "V" or con == "F":
            self.count = self.count + 1
            value.append(lineid[:-5] + "-z-" + "{:03d}".format(self.count))
            value.append(lineid[:-5] + "-N-" + "{:03d}".format(self.count))
            value.append(lineid[:-5] + "-b-" + "{:03d}".format(self.count))
            #value.append(lineid[:-5] + "-btur-" + "{:03d}".format(self.count))
        else:
            value.append(self.con_list[0])
            value.append(self.con_list[1])
            value.append(self.con_list[2])
            #value.append(self.con_list[3])
        self.onButton(2, lineid)
        self.onButton(6, value[0])
        self.onButton(9, value[1])
        self.onButton(12, value[2])
        #self.onButton(15, value[3])
        self.onSelLine()

    def onClinez(self, event):
        lineid = self.value_list[0].GetValue()
        con = self.con_list[0]
        if con == "V" or con == "F":
            self.count = self.count + 1
            value = lineid + "-z-" + "{:03d}".format(self.count)
        else:
            value = con
        self.onButton(6, value)
        self.onSelLine()

    def onCcolden(self, event):
        lineid = self.value_list[0].GetValue()
        con = self.con_list[1]
        if con == "V" or con == "F":
            self.count = self.count + 1
            value = lineid + "-N-" + "{:03d}".format(self.count)
        else:
            value = con
        self.onButton(9, value)

    def onCtherb(self, event):
        lineid = self.value_list[0].GetValue()
        con = self.con_list[2]
        if con == "V" or con == "F":
            self.count = self.count + 1
            value = lineid + "-b-" + "{:03d}".format(self.count)
        else:
            value = con
        self.onButton(12, value)

    def onCturbb(self, event):
        lineid = self.value_list[0].GetValue()
        con = self.con_list[3]
        if con == "V" or con == "F":
            self.count = self.count + 1
            value = lineid + "-btur-" + "{:03d}".format(self.count)
        else:
            value = con
        self.onButton(15, value)

    """
    """
    def onButton(self, col, con):
        last = -1
        found = 0
        while True:
            index = self.cl.GetNextItem(last)
            if index == -1:
                break
            else:
                last = index
                if self.cl.IsSelected(index):
                    if found == 0:
                        found = 1;
                        value = self.value_list[col/3-1].GetValue()
                    if col == 6:
                        wavel = float(self.cl.GetItem(index, 3).GetText())
                        linez = float(self.cl.GetItem(index, 4).GetText())
                        ion_wavel = wavel / (1+linez)
                        self.cl.SetItem(
                            index, 3, "{:4.4f}".format((1+float(value))\
                                                       *ion_wavel))
                    order = int(self.cl.GetItem(index, 1).GetText())
                    if col != 2:
                        self.cl.SetItem(index, col-2, value)
                        self.cl.SetItem(index, col, con)
                        dataPlotManager.editString(order, col-2, value)
                        dataPlotManager.editString(order, col, con)
                    else:
                        ion_id = self.cl.GetItem(index, col).GetText()\
                                                              .rstrip()
                        if ion_id.split("_")[0] != con.split("_")[0]:#[:-5]:

                            self.cl.SetItem(index, col, ion_id)
                            wavel = float(self.cl.GetItem(index, 3).GetText())
                            ion_wavel = float(con[-4:])*0.1
                            linez = wavel/ion_wavel-1
                            linez_str = "{:4.4f}".format(linez)
                            self.cl.SetItem(index, 4, linez_str)
                            dataPlotManager.editString(order, 4, linez_str)
                            self.value_list[1].SetValue("{:4.6f}".format(linez))
                        else:
                            dataPlotManager.editString(order, -2,
                                                       dataPlotManager.systid)

                        dataPlotManager.editString(order, col, ion_id)
    """
    #"""
    """
    def onSelItem(self, event):
        self.onSelLine()

    def onSelLine(self):
        last = -1
        self.sel_sub = list()
        self.sel_vel = list()
        sel_num = 0
        while True:
            index = self.cl.GetNextItem(last)
            if index == -1:
                break
            else:
                last = index
                if self.cl.IsSelected(index):
                    sel_num = sel_num + 1
                    sel_lineid = str(
                        self.cl.GetItem(index, 2).GetText()).strip()
                    sel_wavel = float(self.cl.GetItem(index, 3).GetText())
                    sel_linez = str(
                        self.cl.GetItem(index, 4).GetText()).strip()
                    sel_colden = str(
                        self.cl.GetItem(index, 7).GetText()).strip()
                    sel_therb = str(
                        self.cl.GetItem(index, 10).GetText()).strip()
                    sel_turbb = str(
                        self.cl.GetItem(index, 13).GetText()).strip()
                    sel_linezcon = str(
                        self.cl.GetItem(index, 6).GetText()).strip()
                    sel_coldencon = str(
                        self.cl.GetItem(index, 9).GetText()).strip()
                    sel_therbcon = str(
                        self.cl.GetItem(index, 12).GetText()).strip()
                    sel_turbbcon = str(
                        self.cl.GetItem(index, 15).GetText()).strip()
                    for i in range(0, dataPlotManager.sub_num):
                        sel_vel = (sel_wavel/dataPlotManager.sub_cen[i]-1)\
                                  *299792.458
                        if np.absolute(sel_vel)<dataPlotManager.fact:
                            self.sel_sub.append(i)
                            self.sel_vel.append(sel_vel)

        dataPlotManager.plotProductsGraphics()
        wxApp.figureCanvas.draw()
        if sel_num == 1:
            self.value_all.SetValue(sel_lineid)#[:-5])
            self.value_linez.SetValue(sel_linez)
            self.value_colden.SetValue(sel_colden)
            self.value_therb.SetValue(sel_therb)
            self.value_turbb.SetValue(sel_turbb)
            self.value_list[0].SetValue(sel_lineid)#[:-5])
            self.value_list[1].SetValue(sel_linez)
            self.value_list[2].SetValue(sel_colden)
            self.value_list[3].SetValue(sel_therb)
            self.value_list[4].SetValue(sel_turbb)
            self.con_list = list()
            self.con_list.append(sel_linezcon)
            self.con_list.append(sel_coldencon)
            self.con_list.append(sel_therbcon)
            self.con_list.append(sel_turbbcon)

    def onCheckItem(self, index, flag):
        self.disc[index] = int(flag)
        order = int(self.cl.GetItem(index, 1).GetText())
        if flag is True:
            dataPlotManager.editString(order, 16, "Y")
        else:
            dataPlotManager.editString(order, 16, "N")
        if self.replot is True:
            self.add_line = dataPlotManager.add_line
            self.add_wavel = dataPlotManager.add_wavel
            self.add_norm = dataPlotManager.add_norm
            self.show_line = dataPlotManager.show_line
            dataPlotManager.setPlotablePar()
            dataPlotManager.setPlotableSubSline()
            #dataPlotManager.plotProductsGraphics()
            try:
                dataPlotManager.plotSubSpec(dataPlotManager.sub_sel)
            except:
                dataPlotManager.plotSubSpecs()
            wxApp.figureCanvas.draw()
    """
    """
class LineParManager:
    recipe_name = "espda_fit_line"

    def readFitsData(self, line_table, fitsFiles, edit):
        #self.add_line = add_line
        #self.edit = edit
        self.frames = dict()
        for frame in fitsFiles:
            if frame == '':
                continue
            self.frames[frame.category] = frame

        if "SLINE_VOIGT" in self.frames:
            self.prefix = "SLINE"
        else:
            self.prefix = "FLINE"
        try:
            self.line = LinesFrame(line_table,
                                   self.frames[self.prefix+"_VOIGT"], edit)
        except:
            pass

        #self.path = re.sub("SLINE_" + self.affix + "\.fits", "",
        #    self.frames["SLINE_" + self.affix].name)
        #self.time = datetime.now().time().isoformat
    """
    """
class DataPlotterManager:
    mask_limit  = list()
    recipe_name = "espda_fit_line"
    edit_string = ""

    def setInteractiveParameters(self):
        paramList      = list()
        paramList.append(reflex.RecipeParameter(recipe = self.recipe_name,
            displayName = "hwidth",    group = "Line parameters", description =
            "Halfwidth for line grouping [nm]: [2.5e-3, 10.0]"))
        paramList.append(reflex.RecipeParameter(recipe = self.recipe_name,
            displayName = "par-range", group = "Line parameters", description =
            "Rules to set the range of line parameters\n" \
            "Sample syntax for par-range: '1e-4,10.0,17.0,2.0,50.0,0.0,0.0', " \
            "meaning:\n" \
            "- redshift half-range: 1.0e-4\n" \
            "- minimum/maximum logarithimic column density: 10.0/17.0\n" \
            "- minimum/maximum Doppler broadening: 2.0/50.0\n" \
            "- minimum/maximum turbulence broadening: 0.0/0.0"))
        paramList.append(reflex.RecipeParameter(recipe = self.recipe_name,
            displayName = "edit-file", group = "Line parameters", description =
            "File with instructions to edit line table"))
        paramList.append(reflex.RecipeParameter(recipe = self.recipe_name,
            displayName = "fit-flag",  group = "Fitting",         description =
            "Flag to fit the lines"))
        paramList.append(reflex.RecipeParameter(recipe = self.recipe_name,
            displayName = "iter-max",  group = "Fitting",         description =
            "Maximum number of iterations: [1, 1e3]"))
        paramList.append(reflex.RecipeParameter(recipe = self.recipe_name,
            displayName = "grp-sel",   group = "Fitting",         description =
            "Group of lines selected for fitting: [0, number of groups]"))
        paramList.append(reflex.RecipeParameter(recipe = self.recipe_name,
            displayName="model-resol", group = "Fitting",         description =
            "Resolution of the instrument model [1.0e3, 1.0e6]"))
        return paramList


    def readFitsData(self, fitsFiles):
        self.edit = ""
        self.frames = dict()
        for frame in fitsFiles:
            if frame == '':
                continue
            self.frames[frame.category] = frame
        if "SLINE_VOIGT" in self.frames:
            self.prefix = "SLINE"
            self.resid = "yes"
        else:
            self.prefix = "FLINE"
            self.resid = "no"
        sub = re.sub("FLINE_IDEN.fits", "",
                     self.frames["FLINE_IDEN"].name)
        self.path = "/".join(sub.split("/")[:-2])+"/"
        time = sub.split("/")[-2]
        #self.time = datetime.now().time().isoformat()
        prev = [e[29:-4] for e in os.listdir(self.path) \
                if e[:28]=='edit_'+time]
        #prev = [e[:28] for e in os.listdir(self.path)]
        #self.edit_name = self.path + "edit_" + str(self.time) + ".txt"
        try:
            self.edit_name = self.path+"edit_"+time+"_"+str(int(prev[-1])+1)\
                             +".txt"
        except:
            self.edit_name = self.path+"edit_"+time+"_1.txt"


        #self.fspec_c = self.frames["FSPEC_CONT"]
        self.rspec_c = self.frames["RSPEC_CONT"]
        self.fline_i = self.frames["FLINE_IDEN"]
        self.sline_i = self.frames["SLINE_IDEN"]
        try:
            #self.fspec_f = self.frames["FSPEC_VOIGT"]
            self.rspec_f = self.frames["RSPEC_VOIGT"]
            self.sline_f = self.frames[self.prefix + "_VOIGT"]
            self.exist = True
        except:
            self.exist = False

        self.fline_p = PipelineProduct(self.fline_i)
        if self.exist:
            #self.fspec_p = PipelineProduct(self.fspec_f)
            self.fspec_p = PipelineProduct(self.rspec_f)
            self.rspec_p = PipelineProduct(self.rspec_f)
            self.sline_p = PipelineProduct(self.sline_f)
        else:
            self.fspec_p = PipelineProduct(self.rspec_c)
            self.rspec_p = PipelineProduct(self.rspec_c)
            self.sline_p = PipelineProduct(self.sline_i)

        self.systz  = self.sline_p.readKeyword("HIERARCH ESO DAS SYSTZ")
        self.systid = self.sline_p.readKeyword("HIERARCH ESO QC SYSTID")
        try:
            self.rchisq = self.fspec_p.readKeyword(
                "HIERARCH ESO QC LINE RCHISQ")
        except:
            self.rchisq = -1.0
        self.systid_col  = self.sline_p.readTableColumn(1, "SYSTID")
        self.lineid_col  = self.sline_p.readTableColumn(1, "LINEID")
        self.sline_wavel = self.sline_p.readTableColumn(1, "WAVEL")
        self.order       = self.sline_p.readTableColumn(1, "ORDER")
        self.order_max  = max(self.order)
        self.rspec_wavel = self.rspec_p.readTableColumn(1, "WAVEL")
        self.wavel_min = min(self.rspec_wavel)
        self.wavel_max = max(self.rspec_wavel)
        try:
            self.count = self.sline_p.readTableColumn(1, "LINEZCON")
            self.count = [s[-2:] for s in self.count]
            self.count = [s.replace("  ", "00") for s in self.count]
            self.count = map(int, self.count)
            self.count = max(self.count)
        except:
            self.count = 0

        self.setPlotablePar()
        self.setPlotableData()

    def setPlotablePar(self):
        try:
            self.add_line = lineTabManager.add_line
            self.add_wavel = lineTabManager.add_wavel
            self.add_norm = lineTabManager.add_norm
            self.show_line = lineTabManager.show_line
        except:
            try:
                self.add_line = ionPanelManager.add_line
                self.add_wavel = ionPanelManager.add_wavel
                self.add_norm = ionPanelManager.add_norm
                self.show_line = ionPanelManager.show_line
            except:
                self.add_line = list()
                self.add_wavel = list()
                self.add_norm = list()
                self.show_line = list()

        try:
            self.disc = np.array(lineTabManager.disc, dtype = bool)
            self.disc = self.disc[:len(self.sline_wavel)]
        except:
            self.disc = np.ones(len(self.sline_wavel), dtype = bool)
        try:
            self.add_mask = np.array(lineTabManager.disc, dtype = bool)
            self.add_mask = self.add_mask[len(self.sline_wavel):]
        except:
            self.add_mask = np.ones(len(self.add_line) / 3)

        id_old = ""
        count  = 1
        self.lineid_list  = list()
        self.id_count = list()
        try:
            self.lineid_list = ionPanelManager.lineid_list
        except:
            for i in range(0, len(self.lineid_col)):
                if (self.lineid_col[i] != id_old) \
                   and (self.systid_col[i].strip() == self.systid):
                    id_old = self.lineid_col[i]
                    self.lineid_list.append(self.lineid_col[i])
        for i in range(0, len(self.lineid_col)):
            if (self.lineid_col[i] != id_old) \
               and (self.systid_col[i].strip() == self.systid):
                id_old = self.lineid_col[i]
                self.id_count.append(count)
            elif self.lineid_col[i] == id_old:
                try:
                    self.id_count[-1] = self.id_count[-1] + 1
                except:
                    pass
        try:
            self.sub_num_old = self.sub_num
        except:
            pass
        self.sub_num   = len(self.lineid_list)
        self.sub_systz = []
        self.sub_wavel = []
        self.sub_linez = []
        self.sub_cen   = []
        for i in range(0, self.sub_num):
            self.sub_systz.append(
                self.sline_p.readKeyword("HIERARCH ESO DAS SYSTZ"))
            found = self.sline_p.readTableColumn(1, "LINEID") \
                    == self.lineid_list[i]
            try:
                self.sub_cen.append(ionPanelManager.sub_cen[i])
            except:
                self.sub_wavel.append(self.sline_p.readTableColumn(1, "WAVEL")
                                      [found][0])
                self.sub_linez.append(self.sline_p.readTableColumn(1, "LINEZ")
                                      [found][0])
                self.sub_cen.append((1 + self.sub_systz[i]) \
                                    * self.sub_wavel[i] /\
                                    (1 + self.sub_linez[i]))


    def setPlotableData(self):
        self.rspec = PlotableData(self.rspec_c, y="NORM")
        self.fline = PlotableData(self.fline_i, y="NORM", color="lightblue",
                                  mode="bar")
        if self.exist:
            self.rfit  = PlotableData(self.rspec_f, y="FIT", color="darkgreen")
            self.sline = PlotableData(self.sline_f, y="NORM", color="green",
                                      mode="bar")
        self.sub_rspec = []
        if self.exist:
            self.sub_rfit  = []
            self.sub_fline = []
            self.sub_sline = []

        for i in range(0, self.sub_num):
            self.sub_rspec.append(PlotableData(
                self.rspec_c, x="VEL", y="NORM", cen=self.sub_cen[i]))
            if self.exist:
                self.sub_rfit.append (PlotableData(
                    self.rspec_f, x="VEL", y="FIT", cen=self.sub_cen[i],
                    color="darkgreen", e="RESID"))
                self.sub_fline.append(PlotableData(
                    self.fline_i, x="VEL", y="NORM", cen=self.sub_cen[i],
                    color="lightblue", mode="bar"))
                self.sub_sline.append(PlotableData(
                    self.sline_f, x="VEL", y="NORM", cen=self.sub_cen[i],
                    color="green", mode="bar", disc=self.disc))

    def setPlotableSubSline(self):
        if self.exist:
            self.sub_sline = []
            for i in range(0, self.sub_num):
                self.sub_sline.append(
                    PlotableData(self.sline_f, x="VEL", y="NORM",
                                 cen=self.sub_cen[i], color="green",
                                 mode="bar", disc=self.disc))

    def addSubplots(self, figure):
        self.gs_top = gridspec.GridSpec(1, 1, bottom=0.8)
        self.plot = figure.add_subplot(self.gs_top[0,0])
        col = np.int(np.ceil(self.sub_num / 4.))
        row = min(self.sub_num, 4)
        self.gs_bot = gridspec.GridSpec(row, col, top=0.7, bottom=0.08,
                                        wspace=0.2, hspace=0.15)
        self.sub_plot = []
        for i in range(0, self.sub_num):
            pos_col = np.int(np.ceil((i + 1) / 4.)) - 1
            pos_row = np.mod(i, 4)
            self.sub_plot.append(figure.add_subplot(
                self.gs_bot[pos_row, pos_col]))

    def plotProductsGraphics(self):
        self.plotSpec()
        self.plotSubSpecs()

    def plotSpec(self):
        self.plot.clear()
        tooltip_spec = "Normalized flux v. wavelength.\n" \
            "Grey: non-rebinned data; blue: rebinned data (with errors);\n" \
            "green: line fit, red: residuals.\n" \
            "Click to add lines.\n" \
            "Double click to remove last added lines."
        if self.exist:
            title = "System "+str(self.systid)
            mask = self.add_mask
        else:
            title = "No systems"
            mask = None
        self.rspec.plot(self.plot, title=title, x_label="Wavelength",
                        y_label="Norm. flux", tooltip=tooltip_spec,
                        mask=None)
        self.fline.plot(self.plot, x_min=self.wavel_min, x_max=self.wavel_max,
                        y_min=-0.2, y_max=1.2)
        if self.exist:
            self.rfit.plot(self.plot)
            self.sline.plot(self.plot, x_min=self.wavel_min,
                            x_max=self.wavel_max, y_min=-0.2,
                            y_max=1.2, lines=self.show_line, line_type='show')

    def plotSubSpecs(self):
        for i in range(0, self.sub_num):
            self.plotSubSpec(i)
        if self.exist:
            text = "Line fitting: chi-squared = "\
                   +str(np.around(self.rchisq, decimals=2))
        else:
            text = "Line ranges"
        try:
            self.sub_plot[0].text(0.5, 1.0, text,
                horizontalalignment="center", verticalalignment = "bottom",
                fontsize=12, transform=self.sub_plot[0].transAxes)
        except:
            pass

    def plotSubSpec(self, i):
        tooltip_spec = "Normalized flux v. wavelength.\n" \
            "Grey: non-rebinned data; blue: rebinned data (with errors);\n" \
            "green: line fit, red: residuals.\n" \
            "Click to add lines.\n" \
            "Double click to remove last added lines."
        title_spec = ""
        x_label = "Vel. diff. [km s-1]"
        y_label = "Norm. flux"
        self.sub_plot[i].axes.yaxis.\
            set_ticklabels([0.0,0.2,0.4,0.6,0.8,1.0])
        if i == 0:
            self.fact = 199
        x_min  = -self.fact
        x_max  = +self.fact
        self.sub_plot[i].clear()
        self.sub_rspec[i].plot(self.sub_plot[i], title_spec,
            tooltip_spec, x_label = x_label, y_label = y_label,
            x_min = x_min, x_max = x_max, y_min = -0.19, y_max = 1.2,
            lines=self.add_line, mask=self.add_mask)#self.add_mask)
        if self.exist:
            try:
                for j in range(0, len(lineTabManager.sel_vel)):
                    sel_sub = lineTabManager.sel_sub[j]
                    sel_vel = lineTabManager.sel_vel[j]
                self.sub_plot[sel_sub].axvline(sel_vel, color = "lightblue",
                                               linewidth = 8, alpha = 0.4)
            except:
                pass
            print self.sub_sline[i], self.sub_plot[i]
            self.sub_fline[i].plot(
                self.sub_plot[i], title_spec, tooltip_spec, x_label=x_label,
                y_label=y_label, x_min=x_min, x_max=x_max, y_min=-0.19,
                y_max=1.2)
            self.sub_sline[i].plot(
                self.sub_plot[i], title_spec, tooltip_spec, x_label=x_label,
                y_label=y_label, x_min=x_min, x_max=x_max, y_min=-0.19,
                y_max=1.2)
            self.sub_rfit[i].plot(
                self.sub_plot[i], title_spec, tooltip_spec, x_label=x_label,
                y_label=y_label, x_min=x_min, x_max=x_max, y_min=-0.19,
                y_max=1.2)
        self.sub_plot[i].text(0.0, 1.1, str(self.lineid_list[i]).strip(),
            horizontalalignment = "center", verticalalignment = "center",
            fontsize = 12)


    def plotWidgets(self):
        widgets = list()
        self.clickable_spec = []
        self.vel_arr = list()
        for i in range(0, self.sub_num):
            self.clickable_spec.append(InteractiveClickableSubplot(
                self.sub_plot[i], self.addLine))
            widgets.append(self.clickable_spec[i])
        return widgets
    """
    """
    def addLine(self, point):
        new_flag = 0
        for i in range(0, self.sub_num):
            #new = self.sub_rspec[i].getXPoint(point.xdata)
            new_vel, new_norm = self.sub_rspec[i].getPoint(point.xdata,
                point.ydata)
            self.new_wavel = self.sub_cen[i] * (1 + new_vel / 299792.458)
            cond = point.inaxes.get_yaxis() \
                   == self.sub_plot[i].axes.get_yaxis()
            if cond:
                self.sub_sel = i
                try:
                    self.ion_wavel = ionPanelManager.ion_wavel[i]
                except:
                    try:
                        self.ion_wavel = self.sub_wavel[i]\
                                         /(1+self.sub_linez[i])
                    except:
                        pass
                    #self.ion_wavel = self.sub_wavel[i] / (1 + self.sub_linez[i])
                print self.ion_wavel, self.new_wavel
                wavel_check = np.absolute(self.sline_wavel - self.new_wavel)\
                    < 0.003
                sel_first = 0
                if sum(wavel_check) > 0:
                    try:
                        lineTabManager.onSelLine()
                    except:
                        pass
                    sel_vel = (self.sline_wavel[wavel_check][0] /\
                        self.sub_cen[i] - 1) * 299792.458
                    self.vel_arr.append(sel_vel)
                    self.sub_plot[i].axvline(sel_vel, color = "lightblue",
                        linewidth = 8, alpha = 0.5)
                    self.sub_plot[i].axvline(sel_vel, color = "blue")
                else:
                    self.vel_arr = []
                    new_flag = 1
                    self.add_line.append (self.new_wavel)
                    self.add_line.append (new_norm)
                    self.add_line.append (new_norm)
                    #self.add_vel.append  (new_vel)
                    self.add_wavel.append(self.new_wavel)
                    self.add_norm.append (new_norm)
                    j = len(self.add_line) / 3 - 1

        if new_flag is 1:
            #self.plotSpec()
            for i in range(0, self.sub_num):
                cond = point.inaxes.get_yaxis() \
                       == self.sub_plot[i].axes.get_yaxis()
                if cond:
                    self.lineid_temp = self.lineid_list[i]
            new_params = list()
            new_params.append(reflex.RecipeParameter(recipe = self.recipe_name,
                displayName = "edit-file", value = self.edit_name))
            try:
                on
            except:
                lineTabManager.onAddLine()
            #except:
            #    pass

            return new_params
    """

if __name__ == '__main__':
    import sys
    import wx
    import reflex_interactive_app
    from espda_plot_common import *
    interactive_app = reflex_interactive_app.PipelineInteractiveApp(
        enable_init_sop=True)
    interactive_app.parse_args()
    if not import_success:
        interactive_app.setEnableGUI(False)
    if interactive_app.isGUIEnabled():
        dataPlotManager = DataPlotterManager()
        interactive_app.setPlotManager(dataPlotManager)
        from reflex_interactive_gui import ReflexInteractiveWxApp
        wxApp = ReflexInteractiveWxApp(interactive_app,
                                       interactive_app.dataPlotManager)
        add_start = list()
        for paramWidget in wxApp.shownParamWidgets:
            if paramWidget[1].parameter.displayName == "add":
                add_start = paramWidget[1].paramChangeCtrl.GetValue()
        ionPanelManager = IonPanelManager(wxApp.frame, -1, "Transitions")
        lineTabManager  = LineTabManager (wxApp.frame, -1, "Lines")
        #wxApp.createPlotsArea()
        #wxApp.connectPlottingWidgets()
        wxApp.MainLoop()
    else:
        interactive_app.set_continue_mode()
    interactive_app.print_outputs()
    sys.exit()
