from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
try:
    import numpy as np
    import reflex
    from pipeline_product import PipelineProduct
    import pipeline_display
    import matplotlib.gridspec as gridspec
    import_success = True

except ImportError:
    import_success = False
    print("Error importing modules pyfits, wx, matplotlib, numpy")

def paragraph(text, width=None):
    import textwrap
    if width is None:
        return textwrap.dedent(text).replace('\n', ' ').strip()
    else:
        return textwrap.fill(textwrap.dedent(text), width=width)

class DataPlotterManager:

    recipe_name = "espda_fit_starcont"

    def setInteractiveParameters(self):
        paramList = list()

        paramList.append(reflex.RecipeParameter(recipe = self.recipe_name,
            displayName = "ord-lrejt", group = "Ord", description =
            "Multiplication factor for low rejection threshold:[0., 1.5]. [0.5]"))

        paramList.append(reflex.RecipeParameter(recipe = self.recipe_name,
            displayName = "ord-hrejt",  group = "Ord", description =
            "Multiplication factor for high rejection threshold:[1.5, 3.0]. [2.0]"))

        paramList.append(reflex.RecipeParameter(recipe = self.recipe_name,
            displayName = "ord-niter", group = "Ord", description =
            "Number of iterations of continuum fitting:  [1, 10]. [1]"))

        paramList.append(reflex.RecipeParameter(recipe = self.recipe_name, 
            displayName = "ord-ford410", group = "Ord", description =
            "Order of the fitting function (Wavelength < 410nm):  [1, 9]. [0]"))

        paramList.append(reflex.RecipeParameter(recipe = self.recipe_name,
            displayName = "ord-ford450", group = "Ord", description =
            "Order of the fitting function (Wavelength < 450nm):  [1, 9]. [1]"))
    
        paramList.append(reflex.RecipeParameter(recipe = self.recipe_name,
            displayName = "ord-ford510", group = "Ord", description =
            "Order of the fitting function (Wavelength < 510nm):  [1, 9]. [2]"))

        paramList.append(reflex.RecipeParameter(recipe = self.recipe_name,
            displayName = "ord-ford800", group = "Ord", description =
            "Order of the fitting function (Wavelength < 800nm):  [1, 9]. [3]"))
            
        paramList.append(reflex.RecipeParameter(recipe = self.recipe_name,
            displayName = "flux-kappa",  group = "Flux", description =
            "Threshold for sigma-clipping for rebinning:  [3, 5]. [3]"))

        paramList.append(reflex.RecipeParameter(recipe = self.recipe_name, 
            displayName = "flux-velstep",  group = "Flux", description =
            "Step on velocity binning [km/s]: [0.3, 1.0e2]. [5]"))

        return paramList

    def readFitsData(self, fitsFiles):
        self.wavel_range = list()
        self.frames = dict()
        for frame in fitsFiles:
            if frame == '':
                continue
            category = frame.category
            self.frames[category] = frame
        #self.width = 1
        
        #Selecting different specid to display

        ref_fits = self.frames["NORD_SPEC"]
        self.ref_fits = PipelineProduct(ref_fits)
        self.ref_fits.specid = self.ref_fits.readTableColumn(1, "SPECID")
        
        self.spec = PlotableData(self.frames["NORD_SPEC"],
                                 y = "FLUX", color = "blue",
                                 mode = "scatter",
                                 marker=".", width = 0.5,
                                 #size = 10, ylog = True)
                                 size = 10)
                                 #mode = "plot")
                                 
        self.spec_cont = PlotableData(self.frames["NORD_SPEC"],
                                      y = "CONT", color = "red",
                                      mode = "scatter",
                                      marker=".", width = 0.5,
                                      #size = 10, ylog = True)
                                      size = 10)
                                      #mode = "plot")

                                                  
        self.spec_nord = PlotableData(self.frames["NORD_SPEC"],
                                      ref_fits = self.frames["NORD_SPEC"],
                                      y = "NFLUX", color = "green",
                                      mode = "scatter", marker=".",
                                      width = 0.5, size = 10)
                                      #mode = "plot")
    
        self.spec_merg = PlotableData(self.frames["NMERG_SPEC"],
                                      ref_fits = self.frames["NMERG_SPEC"],
                                      y = "FLUX", color = "blue",
                                      mode = "plot")

    def addSubplots(self, figure):
        gs_spec = gridspec.GridSpec(1, 1, left=0.15, bottom = 0.60, top = 0.95)
        gs_nord = gridspec.GridSpec(1, 1, left=0.15, bottom = 0.35, top = 0.50)
        gs_merg = gridspec.GridSpec(1, 1, left=0.15, bottom = 0.10, top = 0.25)
        self.subplot_spec = figure.add_subplot(gs_spec[0,0])
        self.subplot_nord = figure.add_subplot(gs_nord[0,0])
        self.subplot_merg = figure.add_subplot(gs_merg[0,0])

    def plotProductsGraphics(self):
        self.plotSpec()
        self.firstClick = False

    def plotSpec(self):
        spec_title   = "Spectrum and fit of continuum"
        spec_tooltip = "Blue: order-by-order spectra; red: continuum fit \n"
            #        spec_y_label = r"Flux [erg $\mathdefault{cm^2}$ " \
            #           r"$\mathdefault{s^{-1}}$ $\mathdefault{nm^{-1}}$]"
        spec_y_label = r"Flux [e-]"

        nord_title   = "Normalised order-by-order spectra"
        nord_tooltip = "Green: Order-by-order spectra " \
                       "normalised to continuum fit \n"
        nord_y_label = r"Normalised Flux"

        merg_title    = "Merged spectrum normalised with continuum fit"
        merg_tooltip  = "Blue: Merged normalised spectrum \n"
        merg_y_label  = r"Normalised Flux"

        all_x_label  = "Wavelength [nm]"

        self.fits = PipelineProduct(self.frames["NORD_SPEC"])
        yf = self.fits.readTableColumn(1, "FLUX")
        y_maxi = np.nanmax(yf)
        y_mini = np.nanmin(yf)

        self.spec.plot(self.subplot_spec, title = spec_title,
                           tooltip = spec_tooltip, x_label = all_x_label,
                           y_min = y_mini-0.1*y_maxi, y_max = y_maxi+0.1*y_maxi,
                           y_label = spec_y_label, bars = self.wavel_range)
            
        self.spec_cont.plot(self.subplot_spec, title = spec_title,
                            tooltip = spec_tooltip, x_label = all_x_label,
                            y_min = y_mini-0.1*y_maxi, y_max = y_maxi+0.1*y_maxi,
                            y_label = spec_y_label)

        self.fits = PipelineProduct(self.frames["NORD_SPEC"])
        x = self.fits.readTableColumn(1, "WAVEL")
        y = self.fits.readTableColumn(1, "NFLUX")
        y_max = np.nanmax(y)
        y_min = np.nanmin(y)
        
        im = np.abs(x - 606.8).argmin()
        ip = np.abs(x - 607.4).argmin()
        #print ('index=',im,ip, x[im], x[ip])
        yn = y[im:ip]

        ynmed = np.mean(yn)
        ynstd = np.std(yn)
        ynsnr = ynmed / ynstd

        self.spec_nord.plot(self.subplot_nord, title = nord_title,
                            tooltip = nord_tooltip, x_label = all_x_label,
                            y_min = y_min-0.1, y_max = y_max+0.1,
                            y_label = nord_y_label, bars = self.wavel_range)
            
        #Text plot
        self.subplot_nord.text(0.75, 0.85, "SNR(607nm) = " +
                               str(np.around(ynsnr, decimals=0)),
                               horizontalalignment="right",
                               verticalalignment="bottom",
                               fontsize=8, transform=self.subplot_nord.transAxes)

        self.fits = PipelineProduct(self.frames["NMERG_SPEC"])
        xnm = self.fits.readTableColumn(1, "WAVEL")
        ynm = self.fits.readTableColumn(1, "FLUX")
        
        im = np.abs(xnm - 606.8).argmin()
        ip = np.abs(xnm - 607.4).argmin()
        #print ('index=',im,ip, x[im], x[ip])
        ym = ynm[im:ip]
        ymmed = np.mean(ym)
        ymstd = np.std(ym)
        ymsnr = ymmed / ymstd

        self.spec_merg.plot(self.subplot_merg, title = merg_title,
                            tooltip = merg_tooltip, x_label = all_x_label,
                            y_min=-0.1,y_max=1.2,
                            y_label = merg_y_label, bars = self.wavel_range)
        #Text plot
        self.subplot_merg.text(0.75, 0.90, "SNR(607nm) = " +
                             str(np.around(ymsnr, decimals=0)),
                             horizontalalignment="right",
                             verticalalignment="bottom",
                             fontsize=8, transform=self.subplot_merg.transAxes)

    def plotWidgets(self):
        widgets = list()
        self.clicks = 0
        self.clickable_spec = InteractiveClickableSubplot(self.subplot_spec, 
            self.wavelRange)
        widgets.append(self.clickable_spec)
        return widgets

    def wavelRange(self, point):
        self.clicks = self.clicks + 1
        range_limit = self.spec.getXPoint(point.xdata)
        if self.clicks < 3:
            if range_limit in self.wavel_range:
                self.wavel_range.remove(range_limit)
            else:
                self.wavel_range.append(range_limit)
        self.plotSpec()
        new_params = list()

        return new_params

    def setWindowTitle(self):
      title = 'ESPRESSO Data Analysis Software: Fit Star Continuum' 
      return title

if __name__ == '__main__':
    import reflex_interactive_app
    import sys
    import pkg_resources
    from espda_plot_common import *
    interactive_app = reflex_interactive_app.PipelineInteractiveApp(
        enable_init_sop = True)
    interactive_app.parse_args()
    if not import_success:
        interactive_app.setEnableGUI(False)
    if interactive_app.isGUIEnabled():
        dataPlotManager = DataPlotterManager()
        interactive_app.setPlotManager(dataPlotManager)
        interactive_app.showGUI()
    else:
        interactive_app.set_continue_mode()
    interactive_app.print_outputs()
    sys.exit()
