from __future__ import absolute_import
from __future__ import print_function
import os
import sys
import matplotlib
from matplotlib.figure import Figure
from matplotlib.backends.backend_wxagg import FigureCanvasWxAgg
import wx
from wx import xrc
from wx import lib
from wx.lib import intctrl
from wx.lib import masked
import wx.lib.scrolledpanel
import reflex_navigation_toolbar

# GC.1 - Date and time are needed to keep track of the modifications to the line parameter table 
import datetime
# GC.1

matplotlib.interactive(False)

# This class presents the wx application interface


class ReflexInteractiveWxApp(wx.App):

# GC.2 - The line parameter table manager is required as an additional input
    def __init__(self, interactive_app, dataPlotManager):
#    def __init__(self, interactive_app, dataPlotManager, lineParManager):
# GC.2

        self.inter_app = interactive_app
        self.dataPlotManager = dataPlotManager

# GC.3 - Instance attributes
        """
        self.lineParManager = lineParManager  # Table manager
        self.line_par_toggle = 0              # Table toggling
        self.edit_flag = 0                    # Table editing flag
        self.line = list()
        self.par = list()
        self.value = list()
        """
# GC.3

        # Initialize wx GUI application
        wx.App.__init__(self, 0)

    def OnInit(self):
        'Create the main window and insert the custom frame'
        if hasattr(self.dataPlotManager, 'xrc_file') and \
           self.dataPlotManager.xrc_file != None:
            respath = self.dataPlotManager.xrc_file
        else:
            respath = os.path.join(os.path.dirname(__file__),
                                   'reflex_interactive_gui.xrc')
        self.res = xrc.XmlResource(respath)

        self.frame = self.res.LoadFrame(None, 'ReflexInteractiveWxApp')
        self.frame.SetPosition((-1,0))
        self.setXrcObjects()
        self.bindXrcObjects()
        self.addToolTips()
        self.setFitsFiles()
        self.setWindowTitle()
        self.setDatasetInfoText()
        self.setFilesInfo()
        self.setupParameters()
        self.createPlotsArea()
        self.connectPlottingWidgets()
        self.frame.Show(True)
        return True

    def setXrcObjects(self):
        self.plotPanel = xrc.XRCCTRL(self.frame, 'plotPanel')
        self.ctrlPanel = xrc.XRCCTRL(self.frame, 'ctrlPanel')
        self.ctrlPanel.SetScrollbars( 0, 1,  0, 1 )
        self.statusBar = xrc.XRCCTRL(self.frame, 'statusBar')
        self.parameterNotebook = xrc.XRCCTRL(self.frame, 'parameterNotebook')
        self.parameterNotebook.SetMinSize((170, 262))
        self.datasetInfoText = xrc.XRCCTRL(self.frame, 'datasetInfoText')
        self.setDisableCheck = xrc.XRCCTRL(self.frame, 'setDisableCheck')
        self.setInitSopCheck = xrc.XRCCTRL(self.frame, 'setInitSopCheck')
        self.filesInfo       = xrc.XRCCTRL(self.frame, 'filesInfo')
        self.filesList       = xrc.XRCCTRL(self.frame, 'filesList')
        if(not self.inter_app.is_init_sop_enable()):
            self.setInitSopCheck.Hide()

    def bindXrcObjects(self):
        self.Bind(wx.EVT_BUTTON, self.onCont, id=xrc.XRCID('contBtn'))

# GC.4 - New button to display the line parameter table, defined in the .xrc file
        #self.Bind(wx.EVT_BUTTON, self.onLinePar, id=xrc.XRCID('lineParBtn'))
# GC.4
        self.Bind(wx.EVT_BUTTON, self.onRepeat, id=xrc.XRCID('repeatBtn'))
        self.Bind(wx.EVT_BUTTON, self.onHelp, id=xrc.XRCID('helpBtn'))
        self.Bind(wx.EVT_CHECKBOX, self.onSetDisable, self.setDisableCheck)
        self.Bind(wx.EVT_COLLAPSIBLEPANE_CHANGED, self.onFileInfoPaneChanged)
        self.Bind(wx.EVT_LIST_ITEM_SELECTED, self.onItemSelected)

        if(self.inter_app.is_init_sop_enable()):
            self.Bind(wx.EVT_CHECKBOX, self.onSetInitSop, self.setInitSopCheck)

    def addToolTips(self):
        self.setDisableCheck.SetToolTipString("""This window won't pop up anymore. Take into accoun that the initial values of the parameters will be used (the same ones used in the first loop iteration).""")
        self.setInitSopCheck.SetToolTipString("""The parameters above will be used as initial values (1st iteration) for this recipe loop.""")

    def setWindowTitle(self):
        if not self.frame:
            raise Exception("Frame not found")

        if hasattr(self.dataPlotManager, 'setWindowTitle'):
            self.frame.SetTitle(self.dataPlotManager.setWindowTitle())

    def setupParameters(self):
        # Get the list of recipe parameter groups
        self.requestedParamList = self.dataPlotManager.setInteractiveParameters(
        )
        parameterGroups = list()
        for param in self.requestedParamList:
            if parameterGroups.count(param.group) == 0:
                parameterGroups.append(param.group)

        self.parameterGrid = {}
        self.parameterTabs = {}
        for paramGroup in parameterGroups:
            self.parameterGrid[paramGroup] = wx.FlexGridSizer(cols=2,
                                                              vgap=4, hgap=4)
            self.parameterGrid[paramGroup].AddGrowableCol(1)
            self.parameterTabs[paramGroup] = wx.lib.scrolledpanel.ScrolledPanel(self.parameterNotebook)
            self.parameterTabs[paramGroup].SetupScrolling()
            self.parameterTabs[paramGroup].SetSizer(
                self.parameterGrid[paramGroup], 1)
            self.parameterNotebook.AddPage(
                self.parameterTabs[paramGroup], paramGroup)

        # The widgets for the parameters
        self.shownParam = list()
        self.shownParamWidgets = list()
        for param in self.requestedParamList:
            for param_in_sop in self.inter_app.inputs.in_sop:
                if param_in_sop.recipe == param.recipe and \
                   param_in_sop.displayName == param.displayName:
                    param.value = param_in_sop.value
                    param.name = param_in_sop.name
                    param.valtype = param_in_sop.valtype
                    param.partype = param_in_sop.partype
                    param.valmin = param_in_sop.valmin
                    param.valmax = param_in_sop.valmax
                    param.valenum = param_in_sop.valenum
                    self.shownParam.append(param)
                    paramWidget = ParamWidget(
                        self.parameterTabs[param.group], param)
                    self.shownParamWidgets.append((param.group, paramWidget))

        # Pack parameter widgets
        for (paramGroup, paramWidget) in self.shownParamWidgets:
            self.parameterGrid[paramGroup].Add(paramWidget.paramLabel,
                                               flag=wx.ALL | wx.ALIGN_CENTER_VERTICAL, border=5)
            self.parameterGrid[paramGroup].Add(paramWidget.paramChangeCtrl,
                                               flag=wx.EXPAND | wx.ALL | wx.ALIGN_RIGHT, border=5)
            

    def createPlotsArea(self):
        # Read the pipeline data
        self.dataPlotManager.readFitsData(self.all_fitsFiles)
        # The matplotlib figure
        self.figure = Figure()
        self.figureCanvas = FigureCanvasWxAgg(self.plotPanel, wx.ID_ANY,
                                              self.figure)
#    self.canvas.SetToolTip(wx.ToolTip(tip = ''))
        # Setup toolbar
        self.toolbar = reflex_navigation_toolbar.ReflexNavigationToolbar(
            self.figureCanvas)
        self.frame.SetToolBar(self.toolbar)
        # Bind actions for status bar and tooltips
        self.figureCanvas.mpl_connect('motion_notify_event', self.onMotion)
        self.figureCanvas.mpl_connect('axes_enter_event', self.onEnterAxes)
        self.figureCanvas.mpl_connect('axes_leave_event', self.onLeaveAxes)

        # Bind actions for proper size
        self.plotPanel.Bind(wx.EVT_SIZE, self.onResize)
        self.plotPanel.Bind(wx.EVT_IDLE, self.onIdle)

# GC.A - To be properly displayed on Mac native windowing system, resizing should be performed
        #self.needs_resize = True
# GC.A
        self.dataPlotManager.addSubplots(self.figure)
        self.dataPlotManager.plotProductsGraphics()

# GC.5 - General subplot positioning
        """
        self.figure.subplots_adjust(wspace=0.05, hspace=0.1,
                                    top=0.93, bottom=0.07,
                                    left=0.1, right=0.95)
        """
# GC.5

    def setFitsFiles(self):
        try:
            self.in_sof_fitsFiles = self.inter_app.inputs.in_sof.files
            self.dataset = self.inter_app.inputs.in_sof.datasetName
            self.in_sof_rec_orig_fitsFiles = self.inter_app.inputs.in_sof_rec_orig.files
            self.all_fitsFiles = self.in_sof_fitsFiles + \
                self.in_sof_rec_orig_fitsFiles
            if len(self.in_sof_rec_orig_fitsFiles) == 0:
                self.purposes = 'INVALID'
            else:
                self.purposes = self.in_sof_rec_orig_fitsFiles[0].purposes
        except ValueError:
            print("Error parsing input sof.")
            print("Syntax should be LOSO_NAME|file1;CATEGORY;PURPOSE,file2;...")
            sys.exit()

    def setDatasetInfoText(self):
        # The general info status bar
        datasetInfoAreaText = ''' This data belongs to dataset:
    '''+self.dataset
        self.datasetInfoText.SetLabel(datasetInfoAreaText)

    def setFilesInfo(self):
        #self.filesInfo.Collapse()
        self.filesList.InsertColumn(0, "Category")
        self.filesList.InsertColumn(1, "File Name")
        for file in self.all_fitsFiles :
            self.filesList.Append([file.category, file.name])
        self.filesList.SetColumnWidth(0, wx.LIST_AUTOSIZE)
        self.filesList.SetColumnWidth(1, wx.LIST_AUTOSIZE)
        self.filesList.SetToolTipString("""List of files that have been an input of this window. If you click on a particular file, you can press Ctrl-C to copy the full path name onto the clipboard. """)
        
    def connectPlottingWidgets(self) :
        if hasattr(self.dataPlotManager, 'plotWidgets') :
            self.widgets = self.dataPlotManager.plotWidgets()
            for widget in self.widgets :
                widget.setPostCallback(self.onPlotWidgetEvent) 

    def readEditedParameters(self) :
        # This will update the output parameters based on the entry boxes.
        # It also contains all the parameters that are not shown in the window
        user_edited_param = list()
        for param_in_sop in self.inter_app.inputs.in_sop:
            newParam = param_in_sop
            for (paramGroup, paramWidget) in self.shownParamWidgets:
                if param_in_sop.displayName == paramWidget.parameter.displayName and \
                        param_in_sop.recipe == paramWidget.parameter.recipe:
                    paramWidget.setValue(event=None)
                    newParam = paramWidget.parameter
            user_edited_param.append(newParam)
        self.inter_app.set_edited_param(user_edited_param)

    def onPlotWidgetEvent(self, new_params = None):
        self.figureCanvas.draw()
        if new_params is not None :
            for new_param in new_params :
                for widget in self.shownParamWidgets :
                    if new_param.displayName == widget[1].parameter.displayName :
                        widget[1].paramChangeCtrl.SetValue(new_param.value)
            

    def onItemSelected(self, event):
        currentItem = event.m_itemIndex
        clipData = wx.TextDataObject()
        clipData.SetText(self.all_fitsFiles[self.filesList.GetFocusedItem()].name)
        wx.TheClipboard.Open()
        wx.TheClipboard.SetData(clipData)
        wx.TheClipboard.Close()

    def onFileInfoPaneChanged(self, event):
        self.ctrlPanel.GetSizer().Layout()

    def onCont(self, event):
        self.readEditedParameters()
        self.inter_app.set_continue_mode()
        self.frame.Close(True)

# GC.6 - To continue the workflow, close the line parameter table if it was previously opened
        #if self.line_par_toggle == 1:
        #    self.line_table.Close(True)
# GC.6


# GC.7 - New function to display and interact with the line parameter table
    """
    def onLinePar(self, event):

        # Open window if it is closed
        if self.line_par_toggle == 0:

            # Create table
            self.line_table = wx.Frame(None, -1, title="Edit line parameters", 
                size = (1200,290), pos = (wx.DisplaySize()[0] / 2. - 600, 0))
            self.edit = self.dataPlotManager.edit
            self.lineParManager.readFitsData(self.line_table, 
                self.all_fitsFiles, self.edit) 
            #self.path = self.lineParManager.path
            #self.path = self.dataPlotManager.path
            #self.time = self.dataPlotManager.time
            self.edit_name = self.dataPlotManager.edit_name
            self.grid = self.lineParManager.line.par_grid
            self.hdul = self.grid.hdulist
            self.data = self.hdul[1].data
            line_num = self.grid.line_tot

            # Retrieve new changes and update par_edit
            self.Bind(wx.grid.EVT_GRID_CELL_CHANGE, self.OnChange)   
            self.line_par_toggle = 1

        # Close window if it is open
        else:
            self.line_par_toggle = 0
            self.line_table.Close(True)
    """
# GC.7


# GC.8 - New function to trace the changes to the line parameter table

    """
    def OnChange(self, event):
        self.edit_flag = 1
        #self.time = datetime.datetime.now().time().isoformat()
        #self.line.append(event.GetRow())
        self.line.append(str(self.grid.GetCellValue(event.GetRow(), 13)))
        self.par.append(event.GetCol())
        self.value.append(str(self.grid.GetCellValue(event.GetRow(), 
            event.GetCol())))
        
    
        # When a change is made, the parameter edit-time is updated with 
        # the current system time, to force re-execution also in lazy mode.
        edit_old = str()
        self.edit_file = open(self.edit_name, "a+b")
        edit_old = self.edit_file.read()
        self.edit_file.close
        for widget in self.shownParamWidgets:
            if widget[1].parameter.displayName == "edit-time":
                if edit_old == u"":
                    if widget[1].paramChangeCtrl.GetValue() is not u"":
                        self.edit_old_file = open(widget[1].\
                            paramChangeCtrl.GetValue(), "rb")
                        edit_old = self.edit_old_file.read()
                        self.edit_old_file.close()
                widget[1].paramChangeCtrl.SetValue(self.edit_name)
        for i in range(0, len(self.line)):
            self.edit = edit_old + str(self.line[i]) + ":" + str(self.par[i]) +\
                ":" + str(self.value[i]) + ","
        self.edit_file = open(self.edit_name, "wb")
        self.edit_file.write(self.edit)
        self.edit_file.close()
    """
# GC.8


# GC.9 - On repeat, the changed are saved on the FITS line table frame
    """
    def onRepeat(self, event):

        # This will update the output parameters based on the entry boxes.
        # It also contains all the parameters that are not shown in the window
        user_edited_param = list()
        
        for paramWidget in self.shownParamWidgets:  
            if paramWidget[1].parameter.displayName == "fit-flag":
                if paramWidget[1].paramChangeCtrl.GetValue() == 0:
                    for paramWidget in self.shownParamWidgets:
                        if paramWidget[1].parameter.displayName == "edit-time":
                            paramWidget[1].paramChangeCtrl.SetValue("")
        
        for param_in_sop in self.inter_app.inputs.in_sop:
            newParam = param_in_sop
            for (paramGroup, paramWidget) in self.shownParamWidgets:
                if param_in_sop.displayName == paramWidget.parameter.displayName and \
                        param_in_sop.recipe == paramWidget.parameter.recipe:
                    paramWidget.setValue(event=None)
                    newParam = paramWidget.parameter
            user_edited_param.append(newParam)

        #self.inter_app.set_edited_param(user_edited_param)
        self.inter_app.set_repeat_mode()
        self.readEditedParameters()
        self.frame.Close(True)
        if self.line_par_toggle == 1:
            self.line_table.Close(True)

    """
# GC.9

    def onHelp(self, event):
        window_help_msg = self.dataPlotManager.setWindowHelp()
        general_help_msg = """

The window has several parts:\n
   1. Plot area. This area shows how good the reduction was performed\n
   2. Parameter area. This area shows the parameters used for the execution of the recipe grouped in tabs of related parameters. Depending on the type of parameter a validation is performed that forbids disallowed values (for instance text characters are nor allowed in parameters of type integer). The parameters can be changed if the recipe has to be re-run with new parameters\n
   3. Button area. These buttons control the interactivity of the window\n
        a) Continue wkf. This button sends the current results to the next recipe in the workflow.\n
        b) Re-run recipe. This button will execute the predict recipe again with the new parameters\n
        c) Help. This button shows this help\n
        d) Disable this window... This option will instruct the workflow that the window shouldn't pop up anymore. It has the same effect as setting Enabled=false in the input port of this actor. Take into accoun that the default values of the parameters will be used (the same ones used in the first loop iteration). This feature needs special support from the workflow.\n
        e) Use the parameters above... This will instruct the workflow that the newly entered parameter values are to be used as the initial values of the recipe for the next datasets. This feature needs special support from the workflow.\n
   4. Top toolbar area. These buttons allow certain interactivity with the plots (zoom, shift, layout) as well as exporting to png\n
   5. File into area. The bottom right part of the window contains the list of files that have been an input of this window. Not all the files there have to be neccessarily used for the plots. The CATEGORY is the type of file as produced by the recipe and the File name contains the full path to the file. If you click on a particular file, you can press Ctrl-C to copy the full path name onto the clipboard.
        """
        dlg = wx.MessageDialog(self.frame, window_help_msg + general_help_msg,
                               self.dataPlotManager.setWindowTitle()+" Help",
                               wx.OK | wx.ICON_INFORMATION)
        dlg.ShowModal()
        dlg.Destroy()

    def onSetDisable(self, event):
        self.inter_app.set_next_iteration_enable(not self.setDisableCheck.GetValue())

    def onSetInitSop(self, event):
        self.inter_app.set_next_init_sop(self.setInitSopCheck.GetValue())

    def onMotion(self, event):
        if event.inaxes:
            self.statusBar.SetStatusText(event.inaxes.format_coord(event.xdata,
                                                                   event.ydata))
        else:
            self.statusBar.SetStatusText((''), 0)

    def onEnterAxes(self, event):
        if hasattr(event.inaxes, 'tooltip'):
            tip = self.figureCanvas.GetToolTip()
            if not tip:
                tip = wx.ToolTip(tip=event.inaxes.tooltip)
                self.figureCanvas.SetToolTip(tip)
            else:
                tip.SetTip(event.inaxes.tooltip)
            tip.Enable(True)

    def onLeaveAxes(self, event):
        try:
            self.figureCanvas.GetToolTip().Enable(False)
        except:
            pass

    def onResize(self, event):
        self.needs_resize = True

    def onIdle(self, event):
        if self.needs_resize:
            self.figureCanvas.SetSize(self.plotPanel.GetSize())
            self.needs_resize = False
        wx.WakeUpIdle()


class FloatValidator(wx.PyValidator):
     """ This validator is used to ensure that the user has entered a float
         into the text control of MyFrame.
     """
     def __init__(self):
         wx.PyValidator.__init__(self)
         self.Bind(wx.EVT_CHAR, self.OnChar) 
         self.min = None
         self.max = None

     def Clone(self):
         return FloatValidator()

     def isFloat(self, s) :
         try:
             float(s)
             return True
         except ValueError:
             return False 

     def Validate(self, win):
         return True
 
     def TransferToWindow(self):
         return True 

     def TransferFromWindow(self):
         return True 

     def SetBounds(self, min, max):
         self.min = min
         self.max = max

     def OnChar(self, event):        
         key = event.GetKeyCode()
         ctrl = event.GetEventObject()
         pos = ctrl.GetInsertionPoint()
         textCtrl = self.GetWindow()
         oldText = textCtrl.GetValue()
         selStart, selTo = ctrl.GetSelection()
         selLen = selTo - selStart

         if key < wx.WXK_SPACE or key == wx.WXK_DELETE or key > 255:
             event.Skip()
             return

         if chr(key) == '-':
             # Allow '-' to result in -1 if replacing entire contents:
            if(selLen >= len(oldText) ):
                wx.CallAfter(ctrl.SetValue, "-1")
                wx.CallAfter(ctrl.SetInsertionPoint, 1)
                wx.CallAfter(ctrl.SetSelection, 1, 2)

         newText = oldText[:selStart] + chr(key) + oldText[selTo:]

         if not self.isFloat(newText):
             pass
         else:
             if self.min is not None:
                 if float(newText) < self.min:
                     return
             if self.max is not None:
                 if float(newText) > self.max:
                     return
             event.Skip()

         # Returning without calling even.Skip eats the event before it
         # gets to the text control
         return 

# This class is used to display the label and edit box of a parameter
class ParamWidget:

    def __init__(self, parent, param):
        self.parameter = param
        self.paramLabel = wx.StaticText(parent, label=param.displayName)
        self.paramLabel.SetToolTipString(param.displayName)
        if param.valtype == 'double' :
            if param.partype != 'enum':
                self.paramChangeCtrl = wx.TextCtrl(parent, -1, validator=FloatValidator())
                if param.valmin is not None and param.valmax is not None:
                    self.paramChangeCtrl.GetValidator().SetBounds(param.valmin, param.valmax)
            else :
                choices = []
                for val in param.valenum :
                    choices.append(str(val))
                self.paramChangeCtrl = wx.ComboBox(parent, -1, 
                                       choices = choices, style= wx.CB_READONLY)
            self.paramChangeCtrl.SetValue(str(param.value))
        elif self.parameter.valtype == 'int' :
            if param.partype != 'enum':
                self.paramChangeCtrl = wx.lib.intctrl.IntCtrl(parent)
                if param.valmin is not None and param.valmax is not None:
                    self.paramChangeCtrl.SetBounds(min = param.valmin, max=param.valmax)
                    self.paramChangeCtrl.SetLimited(limited=True)
                self.paramChangeCtrl.SetValue(param.value)
            else :
                choices = []
                for val in param.valenum :
                    choices.append(str(val))
                self.paramChangeCtrl = wx.ComboBox(parent, -1, 
                                       choices = choices, style= wx.CB_READONLY)
                self.paramChangeCtrl.SetValue(str(param.value))
        elif self.parameter.valtype == 'bool' :
            self.paramChangeCtrl = wx.CheckBox(parent)
            self.paramChangeCtrl.SetValue(param.value)
        else:
            if param.partype != 'enum':
                self.paramChangeCtrl = wx.TextCtrl(parent, -1,
                                             style=wx.TE_PROCESS_ENTER)
            else :
                self.paramChangeCtrl = wx.ComboBox(parent, -1,
                                       choices = param.valenum, style= wx.CB_READONLY)
            self.paramChangeCtrl.SetValue(param.value)

        self.paramChangeCtrl.SetToolTipString(param.description)

        # This binding is unnecessary, since the OnRepeat directly access the
        # GetValue() of paramChangeCtrl. I keep it here in case that
        # the desired behaviour is that you have to press Enter to actually edit the
        # value
        self.paramChangeCtrl.Bind(wx.EVT_TEXT_ENTER, self.setValue)

    def setValue(self, event):
        if self.parameter.valtype == 'double' :
            self.parameter.value = float(self.paramChangeCtrl.GetValue())
        elif self.parameter.valtype == 'int' :
            self.parameter.value = int(self.paramChangeCtrl.GetValue())
        else :
            self.parameter.value = self.paramChangeCtrl.GetValue()
