from __future__ import absolute_import
from __future__ import print_function
try:
    import reflex
    import matplotlib
    from pipeline_product import PipelineProduct
    import pipeline_display
    import matplotlib.gridspec as gridspec
    import_success = True
except ImportError:
    import_success = False
    print("Error importing modules pyfits, wx, matplotlib, numpy")

class DataPlotterManager:

    recipe_name = "espda_mask_spec"

    """ General Methods """

    def addSubplots(self, figure):
        self.gs = gridspec.GridSpec(21, 1)
        self.figure = figure
        self.p = Plot()
        self.p.sub_s = self.p.sub(self.figure, self.gs[20,0], xlabel=None,
                                  ylabel=None)


    def plotProductsGraphics(self):
        self.vals = []
        self.bars = []
        self.plotSpec()


    def plotSpec(self):
        self.plotSpecRebin()
        self.plotSpecFull()
        self.plotSpecMask()


    def readFitsData(self, fitsFiles):
        frames = dict()
        for frame in fitsFiles:
            if frame == '':
                continue
            frames[frame.category] = PipelineProduct(frame)

        try:
            self.exp_num = max(frames["FSPEC_PRE"].readTableColumn(1, "SPECID"))
        except:
            self.exp_num = 1
        self.exp_sel = 1
        self.rspec_pre = Data(frames["RSPEC_PRE"])
        self.rspec = Data(frames["RSPEC"])
        try:
            self.fspec_pre = Data(frames["FSPEC_PRE"])
            self.fspec = Data(frames["FSPEC"])
            self.fspec_pre.select("SPECID", 1)
            self.fspec.select("SPECID", 1)
        except:
            self.fspec_pre = Data(frames["RSPEC_PRE"])
            self.fspec = Data(frames["RSPEC"])



    def setInteractiveParameters(self):
        paramList = list()
        paramList.append(reflex.RecipeParameter(
            recipe=self.recipe_name,
            displayName="mask-limit",
            group="Region selection",
            description="Limits of the regions to be masked [nm]. Each pair of "
                        "colon-separated values defines a region. Regions must "
                        "be separated by commas, with a trailing comma after "
                        "the last region."))
        return paramList


    def setWindowTitle(self):
        title = 'ESPRESSO Data Analysis Software: Spectrum Masking'
        return title


    """ Specific methods """

    def defineMask(self, point):
        if point.inaxes.get_yaxis() == self.p.sub_r.axes.get_yaxis():
            exp_sel = 0
            val = self.rspec.get_x(point.xdata)
        if point.inaxes.get_yaxis() == self.p.sub_f.axes.get_yaxis():
            exp_sel = self.exp_sel
            val = self.fspec.get_x(point.xdata)

        if (exp_sel, val) in self.vals:
            self.vals.remove((exp_sel, val))
        else:
            self.vals.append((exp_sel, val))

        for b in self.bars:
            try:
                b.remove()
                self.bars=[]
            except:
                continue

        for w in wxApp.shownParamWidgets:
            if w[1].parameter.displayName == 'mask-limit' and self.first_call:
                self.mask_old = w[1].paramChangeCtrl.GetValue()
                self.first_call = False

        self.plotSpecMask()

        val_str = ""
        for v in self.vals:
            val_str += str(v[0])+":"+str(v[1])+","

        new_params = list()
        new_params.append(reflex.RecipeParameter(
            recipe=self.recipe_name, displayName="mask-limit",
            value=self.mask_old+val_str))
        return new_params


    def plotSpecRebin(self):
        tooltip_r = ("Masked coadded spectrum (frame category: RSPEC), "
                     "overplotted to the unmasked one (frame category: "
                     "RSPEC_PRE). Middle-click on the plot to mask spectral "
                     "regions in the coadded spectrum. Mask limits appear as "
                     "red vertical bars. The regions between each pair of "
                     "consecutive bars, read left to right, will be masked "
                     "when the recipe is re-run.")
        self.p.sub_r = self.p.sub(self.figure, self.gs[:8,0],
                                  title="Masked spectrum", tooltip=tooltip_r)
        self.rspec_pre.step(self.p.sub_r, color='gainsboro')
        self.rspec.step(self.p.sub_r)
        self.p.sub_r.legend(prop={'size': 10})


    def plotSpecFull(self, ret=None):
        tooltip_f = ("Masked spectra of the individual exposures (frame "
                     "category: FSPEC), overplotted to the unmasked ones "
                     "(frame category: FSPEC_PRE). Use the slider below to "
                     "select the exposure to be displayed. Middle-click on the "
                     "plot to mask spectral regions in the selected exposure. "
                     "Mask limits appear as red vertical bars. The regions "
                     "between each pair of consecutive bars, read left to "
                     "right, will be masked when the recipe is re-run.")
        self.p.sub_f = self.p.sub(self.figure, self.gs[10:18,0], ret=ret,
                                  title="Single exposures", tooltip=tooltip_f)
        self.fspec_pre.step(self.p.sub_f, color='gainsboro')
        self.fspec.step(self.p.sub_f, color='gray')
        self.p.sub_f.legend(prop={'size': 10})

    def plotSpecMask(self):
        for v in self.vals:
            if v[0] == 0:
                self.bars.append(self.p.sub_r.axvline(v[1], color='r'))
            elif v[0] == self.exp_sel:
                self.bars.append(self.p.sub_f.axvline(v[1], color='r'))
            else:
                continue

    def plotWidgets(self):
        widgets = list()

        # Clickable plots
        self.first_call = True
        self.click_rspec = InteractiveClickableSubplot(
            self.p.sub_r, self.defineMask)
        widgets.append(self.click_rspec)
        self.click_fspec = InteractiveClickableSubplot(
            self.p.sub_f, self.defineMask)
        widgets.append(self.click_fspec)

        # Slider for exposures
        self.exp_ctrl = Widget().slider_discr(
            self.p.sub_s, self.selectExposure, 1, self.exp_num+1, 1, "Exp.")
        widgets.append(self.exp_ctrl)

        return widgets


    def selectExposure(self, exp_sel):
        self.exp_sel = int(exp_sel)
        self.fspec_pre.select("SPECID", self.exp_sel)
        self.fspec.select("SPECID", self.exp_sel)
        self.p.sub_f.cla()
        self.plotSpecFull(ret=self.p.sub_f)
        self.plotSpecMask()


if __name__ == '__main__':
    import reflex_interactive_app
    import sys
    from espda_plot_common import *
    interactive_app = reflex_interactive_app.PipelineInteractiveApp(
        enable_init_sop=True)
    interactive_app.parse_args()
    if not import_success:
        interactive_app.setEnableGUI(False)
    if interactive_app.isGUIEnabled():
        dataPlotManager = DataPlotterManager()
        interactive_app.setPlotManager(dataPlotManager)
        from reflex_interactive_gui import ReflexInteractiveWxApp
        wxApp = ReflexInteractiveWxApp(interactive_app,
                                       interactive_app.dataPlotManager)
        wxApp.MainLoop()
    else:
        interactive_app.set_continue_mode()
    interactive_app.print_outputs()
    sys.exit()
